import django_filters

from waldur_core.core import filters as core_filters
from waldur_core.structure import filters as structure_filters
from waldur_core.structure import models as structure_models
from waldur_core.structure.filters import ServicePropertySettingsFilter

from . import models


class HostScopeFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return structure_models.ResourceMixin.get_all_models()

    def get_field_name(self):
        return 'scope'


class TriggerFilter(ServicePropertySettingsFilter):
    template = core_filters.URLFilter(view_name='zabbix-template-detail', field_name='template__uuid', distinct=True)
    template_uuid = django_filters.UUIDFilter(field_name='template__uuid')

    class Meta(ServicePropertySettingsFilter.Meta):
        model = models.Trigger
        fields = ServicePropertySettingsFilter.Meta.fields + ('template', 'template_uuid')


class UserFilter(ServicePropertySettingsFilter):
    surname = django_filters.CharFilter(lookup_expr='icontains')
    alias = django_filters.CharFilter(lookup_expr='icontains')

    class Meta(ServicePropertySettingsFilter.Meta):
        model = models.User
        fields = ServicePropertySettingsFilter.Meta.fields + ('alias', 'surname')


class TemplateFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Template


class UserGroupFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.UserGroup



import re
from urllib.parse import urljoin

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker
from model_utils.models import TimeStampedModel

from waldur_core.core import models as core_models
from waldur_core.core.fields import JSONField
from waldur_core.structure import models as structure_models


class JiraService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project, related_name='jira_services', through='JiraServiceProjectLink')

    @classmethod
    def get_url_name(cls):
        return 'jira'


class JiraServiceProjectLink(structure_models.ServiceProjectLink):
    service = models.ForeignKey(on_delete=models.CASCADE, to=JiraService)

    @classmethod
    def get_url_name(cls):
        return 'jira-spl'


class ProjectTemplate(core_models.UiDescribableMixin, structure_models.GeneralServiceProperty):
    @classmethod
    def get_url_name(cls):
        return 'jira-project-templates'

    @classmethod
    def get_backend_fields(cls):
        return super(ProjectTemplate, cls).get_backend_fields() + ('icon_url', 'description')


class Project(structure_models.NewResource, core_models.RuntimeStateMixin):

    class Permissions(structure_models.NewResource.Permissions):
        pass

    service_project_link = models.ForeignKey(
        JiraServiceProjectLink, related_name='projects', on_delete=models.PROTECT)
    template = models.ForeignKey(on_delete=models.CASCADE, to=ProjectTemplate, blank=True, null=True)
    action = models.CharField(max_length=50, blank=True)
    action_details = JSONField(default=dict)

    def get_backend(self):
        return super(Project, self).get_backend(project=self.backend_id)

    def get_access_url(self):
        base_url = self.service_project_link.service.settings.backend_url
        return urljoin(base_url, 'projects/' + self.backend_id)

    @classmethod
    def get_url_name(cls):
        return 'jira-projects'

    @property
    def priorities(self):
        return Priority.objects.filter(settings=self.service_project_link.service.settings)


class JiraPropertyIssue(core_models.UuidMixin, core_models.StateMixin, TimeStampedModel):
    user = models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, null=True)
    backend_id = models.CharField(max_length=255, null=True)

    class Permissions:
        customer_path = 'project__service_project_link__project__customer'
        project_path = 'project__service_project_link__project'

    class Meta:
        abstract = True


class IssueType(core_models.UiDescribableMixin, structure_models.ServiceProperty):
    projects = models.ManyToManyField(Project, related_name='issue_types')
    subtask = models.BooleanField(default=False)

    class Meta(structure_models.ServiceProperty.Meta):
        verbose_name = _('Issue type')
        verbose_name_plural = _('Issue types')

    @classmethod
    def get_url_name(cls):
        return 'jira-issue-types'

    def __str__(self):
        return self.name

    @classmethod
    def get_backend_fields(cls):
        return super(IssueType, cls).get_backend_fields() + (
            'icon_url', 'description', 'subtask', 'projects'
        )


class Priority(core_models.UiDescribableMixin, structure_models.ServiceProperty):

    class Meta(structure_models.ServiceProperty.Meta):
        verbose_name = _('Priority')
        verbose_name_plural = _('Priorities')

    @classmethod
    def get_url_name(cls):
        return 'jira-priorities'

    def __str__(self):
        return self.name

    @classmethod
    def get_backend_fields(cls):
        return super(Priority, cls).get_backend_fields() + ('icon_url', 'description')


class Issue(structure_models.StructureLoggableMixin,
            JiraPropertyIssue):

    type = models.ForeignKey(on_delete=models.CASCADE, to=IssueType)
    parent = models.ForeignKey(on_delete=models.CASCADE, to='Issue', blank=True, null=True)
    project = models.ForeignKey(on_delete=models.CASCADE, to=Project, related_name='issues')
    summary = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    creator_name = models.CharField(blank=True, max_length=255)
    creator_email = models.CharField(blank=True, max_length=255)
    creator_username = models.CharField(blank=True, max_length=255)
    reporter_name = models.CharField(blank=True, max_length=255)
    reporter_email = models.CharField(blank=True, max_length=255)
    reporter_username = models.CharField(blank=True, max_length=255)
    assignee_name = models.CharField(blank=True, max_length=255)
    assignee_email = models.CharField(blank=True, max_length=255)
    assignee_username = models.CharField(blank=True, max_length=255)
    resolution = models.CharField(blank=True, max_length=255)
    resolution_date = models.CharField(blank=True, null=True, max_length=255)
    priority = models.ForeignKey(on_delete=models.CASCADE, to=Priority)
    status = models.CharField(max_length=255)
    updated = models.DateTimeField(auto_now_add=True)

    resource_content_type = models.ForeignKey(on_delete=models.CASCADE, to=ContentType, blank=True, null=True, related_name='jira_issues')
    resource_object_id = models.PositiveIntegerField(blank=True, null=True)
    resource = GenericForeignKey('resource_content_type', 'resource_object_id')

    resolution_sla = models.IntegerField(blank=True, null=True)

    tracker = FieldTracker()

    class Meta:
        unique_together = ('project', 'backend_id')

    def get_backend(self):
        return self.project.get_backend()

    @classmethod
    def get_url_name(cls):
        return 'jira-issues'

    @property
    def key(self):
        return self.backend_id or ''

    @property
    def issue_user(self):
        return self.user  # XXX: avoid logging conflicts

    @property
    def issue_project(self):
        return self.project  # XXX: avoid logging conflicts

    def get_access_url(self):
        base_url = self.project.service_project_link.service.settings.backend_url
        return urljoin(base_url, 'browse/' + (self.backend_id or ''))

    def get_log_fields(self):
        return ('uuid', 'issue_user', 'key', 'summary', 'status', 'issue_project')

    def get_description(self):
        template = settings.WALDUR_JIRA['ISSUE_TEMPLATE']['RESOURCE_INFO']
        if template and self.resource:
            return self.description + template.format(resource=self.resource)

        return self.description

    def __str__(self):
        return '{}: {}'.format(self.uuid.hex, self.backend_id or '???')


class JiraSubPropertyIssue(JiraPropertyIssue):

    class Permissions:
        customer_path = 'issue__project__service_project_link__project__customer'
        project_path = 'issue__project__service_project_link__project'

    class Meta:
        abstract = True


class Comment(structure_models.StructureLoggableMixin,
              JiraSubPropertyIssue):
    issue = models.ForeignKey(on_delete=models.CASCADE, to=Issue, related_name='comments')
    message = models.TextField(blank=True)

    class Meta:
        unique_together = ('issue', 'backend_id')

    def get_backend(self):
        return self.issue.project.get_backend()

    @classmethod
    def get_url_name(cls):
        return 'jira-comments'

    @property
    def comment_user(self):
        return self.user  # XXX: avoid logging conflicts

    def get_log_fields(self):
        return ('uuid', 'comment_user', 'issue')

    def clean_message(self, message):
        template = settings.WALDUR_JIRA['COMMENT_TEMPLATE']
        if not template:
            return self.message

        User = get_user_model()
        template = re.sub(r'([\^~*?:\(\)\[\]|+])', r'\\\1', template)
        pattern = template.format(body='', user=User(full_name=r'(.+?)', username=r'([\w.@+-]+)'))
        match = re.search(pattern, message)

        if match:
            try:
                self.user = User.objects.get(username=match.group(2))
            except User.DoesNotExist:
                pass
            self.message = message[:match.start()]
        else:
            self.message = message

        return self.message

    def prepare_message(self):
        template = settings.WALDUR_JIRA['COMMENT_TEMPLATE']
        if template and self.user:
            return template.format(user=self.user, body=self.message)
        return self.message

    def update_message(self, message):
        self.message = self.clean_message(message)

    def __str__(self):
        return '{}: {}'.format(self.issue.backend_id or '???', self.backend_id or '')


class Attachment(JiraSubPropertyIssue):
    issue = models.ForeignKey(on_delete=models.CASCADE, to=Issue, related_name='attachments')
    file = models.FileField(upload_to='jira_attachments')
    thumbnail = models.FileField(upload_to='jira_attachments_thumbnails', blank=True, null=True)

    class Meta:
        unique_together = ('issue', 'backend_id')

    def get_backend(self):
        return self.issue.project.get_backend()

    @classmethod
    def get_url_name(cls):
        return 'jira-attachments'



from django.core.exceptions import ObjectDoesNotExist
from django.db import migrations


INSTANCE_TYPE = 'OpenStackTenant.Instance'

VOLUME_TYPE = 'OpenStackTenant.Volume'


def set_parent_offering(apps, schema_editor):
    Offering = apps.get_model('marketplace', 'Offering')
    Resource = apps.get_model('marketplace', 'Resource')
    ServiceSettings = apps.get_model('structure', 'ServiceSettings')
    Tenant = apps.get_model('openstack', 'Tenant')
    ContentType = apps.get_model('contenttypes', 'ContentType')
    content_type = ContentType.objects.get_for_model(Tenant)

    for offering in Offering.objects.filter(type__in=(INSTANCE_TYPE, VOLUME_TYPE)).all():
        service_settings_id = offering.object_id
        if not service_settings_id:
            continue

        try:
            service_settings = ServiceSettings.objects.get(id=service_settings_id)
        except ObjectDoesNotExist:
            continue

        tenant_id = service_settings.object_id
        if not tenant_id:
            continue

        try:
            resource = Resource.objects.get(object_id=tenant_id, content_type=content_type)
        except ObjectDoesNotExist:
            continue

        if not offering.parent:
            offering.parent = resource.offering
            offering.save(update_fields=['parent'])


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace_openstack', '0003_set_storage_mode'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('openstack', '0002_volumetype'),
    ]

    operations = [
        migrations.RunPython(set_parent_offering)
    ]



import logging

from django.contrib.contenttypes.models import ContentType

from waldur_core.core import utils as core_utils
from waldur_mastermind.common import mixins as common_mixins
from waldur_mastermind.invoices import models as invoice_models
from waldur_mastermind.invoices import registrators
from waldur_mastermind.marketplace import models as marketplace_models


logger = logging.getLogger(__name__)


def component_usage_register(component_usage):
    from waldur_mastermind.slurm_invoices.registrators import AllocationRegistrator

    plan_period = component_usage.plan_period
    if not plan_period:
        logger.warning('Skipping processing of component usage with ID %s because '
                       'plan period is not defined.', component_usage.id)
        return

    try:
        plan = plan_period.plan
        plan_component = plan.components.get(component=component_usage.component)
        allocation = component_usage.resource.scope
        customer = allocation.project.customer
        invoice, created = registrators.RegistrationManager.get_or_create_invoice(customer, component_usage.date)

        details = AllocationRegistrator().get_component_details(allocation, plan_component)
        details['plan_period_id'] = plan_period.id
        offering_component = plan_component.component

        month_start = core_utils.month_start(component_usage.date)
        month_end = core_utils.month_end(component_usage.date)

        start = month_start if not component_usage.plan_period.start else \
            max(component_usage.plan_period.start, month_start)
        end = month_end if not component_usage.plan_period.end else \
            min(component_usage.plan_period.end, month_end)

        invoice_models.InvoiceItem.objects.create(
            content_type=ContentType.objects.get_for_model(allocation),
            object_id=allocation.id,
            project=allocation.project,
            invoice=invoice,
            start=start,
            end=end,
            details=details,
            unit_price=plan_component.price,
            quantity=component_usage.usage,
            unit=common_mixins.UnitPriceMixin.Units.QUANTITY,
            product_code=offering_component.product_code or plan.product_code,
            article_code=offering_component.article_code or plan.article_code,
        )

    except marketplace_models.PlanComponent.DoesNotExist:
        logger.warning('Plan component for usage component %s is not found.', component_usage.id)



from django.contrib import admin

from waldur_core.structure import admin as structure_admin

from .models import DigitalOceanService, DigitalOceanServiceProjectLink, Droplet


admin.site.register(Droplet, structure_admin.VirtualMachineAdmin)
admin.site.register(DigitalOceanService, structure_admin.ServiceAdmin)
admin.site.register(DigitalOceanServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)



from rest_framework import serializers

from . import models


class AuthResultSerializer(serializers.ModelSerializer):

    token = serializers.SerializerMethodField()

    class Meta:
        model = models.AuthResult
        fields = ('uuid', 'token', 'phone', 'message', 'state', 'error_message', 'details')
        write_only_fields = ('phone', )
        read_only_fields = ('uuid', 'token', 'message', 'state', 'error_message', 'details')

    def get_token(self, auth_result):
        if auth_result.user:
            return auth_result.user.auth_token.key
        return None



from django.core.exceptions import ObjectDoesNotExist
from rest_framework import serializers
from rest_framework.reverse import reverse

from waldur_core.core import utils as core_utils
from waldur_mastermind.marketplace import processors
from waldur_mastermind.marketplace.utils import get_order_item_url
from waldur_mastermind.marketplace_support.utils import format_description
from waldur_mastermind.support import models as support_models

from .views import IssueViewSet, OfferingViewSet


class CreateRequestProcessor(processors.BaseCreateResourceProcessor):
    viewset = OfferingViewSet

    def get_post_data(self):
        order_item = self.order_item
        try:
            template = order_item.offering.scope
        except ObjectDoesNotExist:
            template = None

        if not isinstance(template, support_models.OfferingTemplate):
            raise serializers.ValidationError('Offering has invalid scope. Support template is expected.')

        project = order_item.order.project
        project_url = reverse('project-detail', kwargs={'uuid': project.uuid.hex})
        template_url = reverse('support-offering-template-detail', kwargs={'uuid': template.uuid.hex})
        attributes = order_item.attributes.copy()

        post_data = dict(
            project=project_url,
            template=template_url,
            name=attributes.pop('name', ''),
            order_item=core_utils.serialize_instance(order_item),
        )

        description = attributes.pop('description', '')
        description += format_description('CREATE_RESOURCE_TEMPLATE', {
            'order_item': order_item,
            'order_item_url': get_order_item_url(order_item),
        })

        if order_item.limits:
            components_map = order_item.offering.get_usage_components()
            for key, value in order_item.limits.items():
                component = components_map.get(key)
                if component:
                    description += "\n%s (%s): %s %s" % \
                                   (component.name, component.type, value, component.measured_unit)

        if order_item.plan and order_item.plan.scope:
            post_data['plan'] = reverse('support-offering-plan-detail', kwargs={
                'uuid': order_item.plan.scope.uuid
            })

        if description:
            post_data['description'] = description
        if attributes:
            post_data['attributes'] = attributes
        return post_data


class DeleteRequestProcessor(processors.DeleteResourceProcessor):
    def get_viewset(self):
        return IssueViewSet

    def get_resource(self):
        return self.order_item


class UpdateRequestProcessor(processors.UpdateResourceProcessor):
    def get_view(self):
        return IssueViewSet.as_view({'post': 'update'})

    def get_post_data(self):
        return {'uuid': str(self.order_item.uuid)}



from celery import shared_task
import subprocess  # nosec


@shared_task(name='waldur_auth_saml2.sync_providers')
def sync_providers():
    # It is assumed that waldur console script is installed
    command = ['waldur', 'sync_saml2_providers']
    subprocess.check_output(command, stderr=subprocess.STDOUT)  # nosec



# Generated by Django 2.2.7.dev20191011095232 on 2019-10-30 14:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0001_squashed_0093'),
    ]

    operations = [
        migrations.AddField(
            model_name='offeringcomponent',
            name='use_limit_for_billing',
            field=models.BooleanField(default=False, help_text='Charge for usage-based component is based on user-requested limit.'),
        ),
    ]



# The dancing with the function and its deletion is done
# to keep the namespace clean: only __version__ is going to be exposed.


def _get_version(package_name='waldur_mastermind'):
    import pkg_resources

    # Based on http://stackoverflow.com/a/17638236/175349
    # and https://github.com/pwaller/__autoversion__/blob/master/__autoversion__.py

    try:
        return pkg_resources.get_distribution(package_name).version
    except pkg_resources.DistributionNotFound:
        import os.path
        import re
        import subprocess  # nosec

        repo_dir = os.path.join(os.path.dirname(__file__), os.path.pardir)

        try:
            with open(os.devnull, 'w') as DEV_NULL:
                description = subprocess.check_output(   # nosec
                    ['git', 'describe', '--tags', '--dirty=.dirty'],
                    cwd=repo_dir, stderr=DEV_NULL
                ).strip()

            v = re.search(r'-[0-9]+-', description)
            if v is not None:
                # Replace -n- with -branchname-n-
                # branch = r"-{0}-\1-".format(cls.get_branch(path))
                description, _ = re.subn('-([0-9]+)-', r'+\1.', description, 1)

            if description[0] == 'v':
                description = description[1:]

            return description
        except (OSError, subprocess.CalledProcessError):
            return 'unknown'


__version__ = _get_version()



import re

from django.core.validators import RegexValidator
from django.utils.translation import ugettext_lazy as _

from waldur_openstack.openstack_tenant.views import InstanceViewSet


ClusterNameValidator = RegexValidator(
    regex=re.compile(r"^[a-z0-9]([-a-z0-9])+[a-z0-9]$"),
    message=_('Name must consist of lower case alphanumeric characters or \'-\', '
              'and must start and end with an alphanumeric character')
)


def related_vm_can_be_deleted(node):
    validators = InstanceViewSet.destroy_validators

    for validator in validators:
        if node.instance:
            validator(node.instance)


def all_cluster_related_vms_can_be_deleted(cluster):
    for node in cluster.node_set.all():
        related_vm_can_be_deleted(node)



from waldur_core.core import WaldurExtension


class BookingExtension(WaldurExtension):
    class Settings:
        pass

    @staticmethod
    def django_app():
        return 'waldur_mastermind.booking'

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        from celery.schedules import crontab
        return {
            'waldur-mastermind-send-notifications-about-upcoming-bookings': {
                'task': 'waldur_mastermind.booking.send_notifications_about_upcoming_bookings',
                'schedule': crontab(minute=0, hour=9),
                'args': (),
            }
        }



import django_filters

from . import models


class IdentityProviderFilter(django_filters.FilterSet):
    class Meta:
        model = models.IdentityProvider
        fields = ('name',)

    name = django_filters.CharFilter(lookup_expr='icontains')



from django.apps import AppConfig
from django.db.models import signals


class PlaybookJobsConfig(AppConfig):
    name = 'waldur_ansible.playbook_jobs'
    verbose_name = 'Waldur Ansible Playbooks'

    def ready(self):
        from . import handlers

        Playbook = self.get_model('Playbook')

        signals.pre_delete.connect(
            handlers.delete_playbook_workspace,
            sender=Playbook,
            dispatch_uid='waldur_ansible.handlers.delete_playbook_workspace',
        )

        signals.pre_save.connect(
            handlers.resize_playbook_image,
            sender=Playbook,
            dispatch_uid='waldur_ansible.handlers.resize_playbook_image',
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-04 16:26
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_azure', '0010_sql_database'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sqldatabase',
            name='charset',
            field=models.CharField(blank=True, default='utf8', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='sqldatabase',
            name='collation',
            field=models.CharField(blank=True, default='utf8_general_ci', max_length=255, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2018-01-30 13:48
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_zabbix', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='trigger',
            name='priority',
            field=models.IntegerField(default=0),
        ),
    ]



from django.apps import AppConfig
from django.contrib.auth import get_user_model


class SAML2Config(AppConfig):
    name = 'waldur_auth_saml2'
    verbose_name = 'Auth SAML2'

    def ready(self):
        from djangosaml2.signals import pre_user_save
        from . import handlers

        pre_user_save.connect(
            handlers.update_registration_method,
            sender=get_user_model(),
            dispatch_uid='waldur_auth_saml2.handlers.update_registration_method',
        )



from waldur_core.core import executors as core_executors, tasks as core_tasks
from waldur_core.core import utils as core_utils
from waldur_core.structure import executors as structure_executors

from . import models, tasks


class IssueDeleteExecutor(core_executors.DeleteExecutorMixin, core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, issue, serialized_issue, **kwargs):
        return core_tasks.BackendMethodTask().si(serialized_issue, 'delete_issue'),


class SupportCleanupExecutor(structure_executors.BaseCleanupExecutor):
    pre_models = (
        models.Offering,
    )

    executors = (
        (models.Issue, IssueDeleteExecutor),
    )


class OfferingIssueCreateExecutor(core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, offering, serialized_offering, **kwargs):
        serialized_issue = core_utils.serialize_instance(offering.issue)
        return tasks.create_issue.si(serialized_issue)



from django.apps import AppConfig
from django_fsm import signals as fsm_signals


class GeoIPConfig(AppConfig):
    name = 'waldur_geo_ip'

    def ready(self):
        from waldur_geo_ip.mixins import IPCoordinatesMixin
        from . import handlers

        for index, model in enumerate(IPCoordinatesMixin.get_all_models()):
            fsm_signals.post_transition.connect(
                handlers.detect_vm_coordinates,
                sender=model,
                dispatch_uid='waldur_geo_ip.handlers.detect_vm_coordinates_{}_{}'.format(
                    model.__name__, index),
            )



import copy
import json
import uuid

from django.core.exceptions import ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.utils.encoding import smart_text
from django.utils.translation import ugettext_lazy as _
import pycountry
from rest_framework import serializers

from waldur_core.core import utils, validators as core_validators
from waldur_core.core.validators import validate_cron_schedule


class CronScheduleField(models.CharField):
    description = "A cron schedule in textual form"

    def __init__(self, *args, **kwargs):
        kwargs['validators'] = [validate_cron_schedule] + kwargs.get('validators', [])
        kwargs['max_length'] = kwargs.get('max_length', 15)
        super(CronScheduleField, self).__init__(*args, **kwargs)


class MappedChoiceField(serializers.ChoiceField):
    """
    A choice field that maps enum values from representation to model ones and back.

    :Example:

    >>> # models.py
    >>> class IceCream(models.Model):
    >>>     class Meta:
    >>>         app_label = 'myapp'
    >>>
    >>>     CHOCOLATE = 0
    >>>     VANILLA = 1
    >>>
    >>>     FLAVOR_CHOICES = (
    >>>         (CHOCOLATE, _('Chocolate')),
    >>>         (VANILLA, _('Vanilla')),
    >>>     )
    >>>
    >>>     flavor = models.SmallIntegerField(choices=FLAVOR_CHOICES)
    >>>
    >>> # serializers.py
    >>> class IceCreamSerializer(serializers.ModelSerializer):
    >>>     class Meta:
    >>>         model = IceCream
    >>>
    >>>     flavor = MappedChoiceField(
    >>>         choices={
    >>>             'chocolate': _('Chocolate'),
    >>>             'vanilla': _('Vanilla'),
    >>>         },
    >>>         choice_mappings={
    >>>             'chocolate': IceCream.CHOCOLATE,
    >>>             'vanilla': IceCream.VANILLA,
    >>>         },
    >>>     )
    >>>
    >>> model1 = IceCream(flavor=IceCream.CHOCOLATE)
    >>> serializer1 = IceCreamSerializer(instance=model1)
    >>> serializer1.data
    {'flavor': 'chocolate', u'id': None}
    >>>
    >>> data2 = {'flavor': 'vanilla'}
    >>> serializer2 = IceCreamSerializer(data=data2)
    >>> serializer2.is_valid()
    True
    >>> serializer2.validated_data["flavor"] == IceCream.VANILLA
    True
    """

    def __init__(self, choice_mappings, **kwargs):
        super(MappedChoiceField, self).__init__(**kwargs)

        assert set(self.choices.keys()) == set(choice_mappings.keys()), 'Choices do not match mappings'
        assert len(set(choice_mappings.values())) == len(choice_mappings), 'Mappings are not unique'

        self.mapped_to_model = choice_mappings
        self.model_to_mapped = {v: k for k, v in choice_mappings.items()}

    def to_internal_value(self, data):
        if data == '' and self.allow_blank:
            return ''

        data = super(MappedChoiceField, self).to_internal_value(data)

        try:
            return self.mapped_to_model[str(data)]
        except KeyError:
            self.fail('invalid_choice', input=data)

    def to_representation(self, value):
        if value in ('', None):
            return value

        value = self.model_to_mapped[value]

        return super(MappedChoiceField, self).to_representation(value)


class NaturalChoiceField(MappedChoiceField):
    def __init__(self, choices=None, **kwargs):
        super(NaturalChoiceField, self).__init__(
            choices=[(v, v) for k, v in choices],
            choice_mappings={v: k for k, v in choices},
            **kwargs)


class TimestampField(serializers.Field):
    """
    Unix timestamp field mapped to datetime object.
    """

    def to_representation(self, value):
        return utils.datetime_to_timestamp(value)

    def to_internal_value(self, value):
        try:
            return utils.timestamp_to_datetime(value)
        except ValueError:
            raise serializers.ValidationError(_('Value "%s" should be valid UNIX timestamp.') % value)


class CountryField(models.CharField):

    COUNTRIES = [(country.alpha2, country.name) for country in pycountry.countries]

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 2)
        kwargs.setdefault('choices', CountryField.COUNTRIES)

        super(CountryField, self).__init__(*args, **kwargs)


class StringUUID(uuid.UUID):
    """
    Default UUID class __str__ method returns hyphenated string.
    This class returns non-hyphenated string.
    """

    def __unicode__(self):
        return str(str(self))

    def __str__(self):
        return self.hex

    def __len__(self):
        return len(self.__unicode__())


class UUIDField(models.UUIDField):
    """
    This class implements backward-compatible non-hyphenated rendering of UUID values.
    Default field parameters are not exposed in migrations.
    """

    def __init__(self, **kwargs):
        kwargs['default'] = lambda: StringUUID(uuid.uuid4().hex)
        kwargs['editable'] = False
        kwargs['unique'] = True
        super(UUIDField, self).__init__(**kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(UUIDField, self).deconstruct()
        del kwargs['default']
        del kwargs['editable']
        del kwargs['unique']
        return name, path, args, kwargs

    def _parse_uuid(self, value):
        if not value:
            return None
        return StringUUID(smart_text(value))

    def from_db_value(self, value, expression, connection, context):
        return self._parse_uuid(value)

    def to_python(self, value):
        return self._parse_uuid(value)


class BackendURLField(models.URLField):
    default_validators = [core_validators.BackendURLValidator()]


class JSONField(models.TextField):
    def __init__(self, *args, **kwargs):
        self.dump_kwargs = kwargs.pop('dump_kwargs', {
            'cls': DjangoJSONEncoder,
            'separators': (',', ':')
        })
        self.load_kwargs = kwargs.pop('load_kwargs', {})

        super(JSONField, self).__init__(*args, **kwargs)

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)

    def to_python(self, value):
        if isinstance(value, str) and value:
            try:
                return json.loads(value, **self.load_kwargs)
            except ValueError:
                raise ValidationError(_('Enter valid JSON'))
        return value

    def get_prep_value(self, value):
        """Convert JSON object to a string"""
        if self.null and value is None:
            return None
        return json.dumps(value, **self.dump_kwargs)

    def get_default(self):
        """
        Returns the default value for this field.
        The default implementation on models.Field calls force_unicode
        on the default, which means you can't set arbitrary Python
        objects as the default. To fix this, we just return the value
        without calling force_unicode on it. Note that if you set a
        callable as a default, the field will still call it. It will
        *not* try to pickle and encode it.
        """
        if self.has_default():
            if callable(self.default):
                return self.default()
            return copy.deepcopy(self.default)
        # If the field doesn't have a default, then we punt to models.Field.
        return super(JSONField, self).get_default()



from django.contrib.admin.filters import (
    AllValuesFieldListFilter,
    ChoicesFieldListFilter,
    RelatedFieldListFilter, RelatedOnlyFieldListFilter
)


class DropdownFilter(AllValuesFieldListFilter):
    template = 'admin/core/dropdown_filter.html'


class ChoiceDropdownFilter(ChoicesFieldListFilter):
    template = 'admin/core/dropdown_filter.html'


class RelatedDropdownFilter(RelatedFieldListFilter):
    template = 'admin/core/dropdown_filter.html'


class RelatedOnlyDropdownFilter(RelatedOnlyFieldListFilter):
    template = 'admin/core/dropdown_filter.html'



# -*- coding: utf-8 -*-
from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('openstack', '0001_squashed_0042'),
        ('structure', '0001_squashed_0054'),
    ]

    operations = [
        migrations.CreateModel(
            name='PackageComponent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('ram', 'RAM, MB'), ('cores', 'Cores'), ('storage', 'Storage, MB')], max_length=50)),
                ('amount', models.PositiveIntegerField(default=0)),
                ('price', models.DecimalField(decimal_places=7, default=0, help_text='The price per unit of amount', max_digits=13, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price per hour')),
            ],
        ),
        migrations.CreateModel(
            name='PackageTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('icon_url', models.URLField(blank=True, verbose_name='icon url')),
                ('uuid', waldur_core.core.fields.UUIDField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='packagecomponent',
            name='template',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='components', to='packages.PackageTemplate'),
        ),
        migrations.AlterField(
            model_name='packagecomponent',
            name='price',
            field=models.DecimalField(decimal_places=10, default=0, max_digits=14, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price per unit per day'),
        ),
        migrations.AlterUniqueTogether(
            name='packagecomponent',
            unique_together=set([('type', 'template')]),
        ),
        migrations.CreateModel(
            name='OpenStackPackage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('service_settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='packagetemplate',
            name='service_settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='openstackpackage',
            name='template',
            field=models.ForeignKey(help_text='Tenant will be created based on this template.', on_delete=django.db.models.deletion.CASCADE, related_name='openstack_packages', to='packages.PackageTemplate'),
        ),
        migrations.AddField(
            model_name='openstackpackage',
            name='tenant',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='openstack.Tenant'),
        ),
        migrations.AlterModelOptions(
            name='openstackpackage',
            options={'verbose_name': 'OpenStack VPC package', 'verbose_name_plural': 'OpenStack VPC packages'},
        ),
        migrations.AlterModelOptions(
            name='packagetemplate',
            options={'verbose_name': 'VPC package template', 'verbose_name_plural': 'VPC package templates'},
        ),
        migrations.AddField(
            model_name='packagetemplate',
            name='category',
            field=models.CharField(choices=[('small', 'Small'), ('medium', 'Medium'), ('large', 'Large'), ('trial', 'Trial')], default='small', max_length=10),
        ),
        migrations.AlterField(
            model_name='openstackpackage',
            name='service_settings',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='structure.ServiceSettings'),
        ),
        migrations.AlterField(
            model_name='packagecomponent',
            name='type',
            field=models.CharField(choices=[('ram', 'RAM'), ('cores', 'Cores'), ('storage', 'Storage')], max_length=50),
        ),
        migrations.AddField(
            model_name='packagetemplate',
            name='archived',
            field=models.BooleanField(default=False, help_text='Forbids creation of new packages.'),
        ),
        migrations.AlterField(
            model_name='openstackpackage',
            name='template',
            field=models.ForeignKey(help_text='Tenant will be created based on this template.', on_delete=django.db.models.deletion.PROTECT, related_name='openstack_packages', to='packages.PackageTemplate'),
        ),
        migrations.AddField(
            model_name='packagetemplate',
            name='product_code',
            field=models.CharField(blank=True, max_length=30),
        ),
        migrations.AddField(
            model_name='packagetemplate',
            name='article_code',
            field=models.CharField(blank=True, max_length=30),
        ),
        migrations.AlterField(
            model_name='openstackpackage',
            name='tenant',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='openstack.Tenant'),
        ),
        migrations.AddField(
            model_name='packagetemplate',
            name='unit',
            field=models.CharField(choices=[('month', 'Per month'), ('half_month', 'Per half month'), ('day', 'Per day'), ('hour', 'Per hour'), ('quantity', 'Quantity')], default='day', max_length=30),
        ),
        migrations.AlterField(
            model_name='packagecomponent',
            name='price',
            field=models.DecimalField(decimal_places=10, default=0, max_digits=14, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price per unit'),
        ),
    ]



# Generated by Django 2.2.7 on 2019-12-04 09:18

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0007_offering_enable_dynamic_components'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='offering',
            name='enable_dynamic_components',
        ),
    ]



from django.db import IntegrityError
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers
from rest_framework.reverse import reverse

from waldur_core.core.serializers import GenericRelatedField, AugmentedSerializerMixin
from waldur_core.cost_tracking import models
from waldur_core.structure import SupportedServices, models as structure_models
from waldur_core.structure.filters import ScopeTypeFilterBackend


class PriceEstimateSerializer(AugmentedSerializerMixin, serializers.HyperlinkedModelSerializer):
    scope = GenericRelatedField(related_models=models.PriceEstimate.get_estimated_models())
    scope_name = serializers.SerializerMethodField()
    scope_type = serializers.SerializerMethodField()
    resource_type = serializers.SerializerMethodField()
    consumption_details = serializers.SerializerMethodField(
        help_text=_('How much of each consumables were used by resource.'))

    def __init__(self, *args, **kwargs):
        depth = kwargs.get('context', {}).pop('depth', 0)  # allow to modify depth dynamically
        self.Meta.depth = depth
        super(PriceEstimateSerializer, self).__init__(*args, **kwargs)

    class Meta:
        model = models.PriceEstimate
        fields = ('url', 'uuid', 'scope', 'total', 'consumed', 'month', 'year',
                  'scope_name', 'scope_type', 'resource_type', 'consumption_details', 'children')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_fields(self):
        display_children = self.Meta.depth > 0
        fields = super(PriceEstimateSerializer, self).get_fields()
        if not display_children:
            del fields['children']
        return fields

    def build_nested_field(self, field_name, relation_info, nested_depth):
        """ Use PriceEstimateSerializer to serialize estimate children """
        if field_name != 'children':
            return super(PriceEstimateSerializer, self).build_nested_field(field_name, relation_info, nested_depth)
        field_class = self.__class__
        field_kwargs = {'read_only': True, 'many': True, 'context': {'depth': nested_depth - 1}}
        return field_class, field_kwargs

    def get_scope_name(self, obj):
        if obj.scope:
            return getattr(obj.scope, 'name', str(obj.scope))
        if obj.details:
            return obj.details.get('scope_name')

    def get_scope_type(self, obj):
        return ScopeTypeFilterBackend.get_scope_type(obj.content_type.model_class())

    def get_resource_type(self, obj):
        if obj.is_resource_estimate():
            return SupportedServices.get_name_for_model(obj.content_type.model_class())

    def get_consumption_details(self, obj):
        try:
            consumption_details = obj.consumption_details
        except models.ConsumptionDetails.DoesNotExist:
            return
        consumed_in_month = consumption_details.consumed_in_month
        consumable_items = consumed_in_month.keys()
        pretty_names = models.DefaultPriceListItem.get_consumable_items_pretty_names(
            obj.content_type, consumable_items)
        return {pretty_names[item]: consumed_in_month[item] for item in consumable_items}


class YearMonthField(serializers.CharField):
    """ Field that support year-month representation in format YYYY.MM """

    def to_internal_value(self, value):
        try:
            year, month = [int(el) for el in value.split('.')]
        except ValueError:
            raise serializers.ValidationError(_('Value "%s" should be valid be in format YYYY.MM') % value)
        if not 0 < month < 13:
            raise serializers.ValidationError(_('Month has to be from 1 to 12.'))
        return year, month


class PriceEstimateDateFilterSerializer(serializers.Serializer):
    date_list = serializers.ListField(
        child=YearMonthField(),
        required=False
    )


class PriceEstimateDateRangeFilterSerializer(serializers.Serializer):
    start = YearMonthField(required=False)
    end = YearMonthField(required=False)

    def validate(self, data):
        if 'start' in data and 'end' in data and data['start'] >= data['end']:
            raise serializers.ValidationError(_('Start has to be earlier than end.'))
        return data


class PriceListItemSerializer(AugmentedSerializerMixin,
                              serializers.HyperlinkedModelSerializer):
    service = GenericRelatedField(related_models=structure_models.Service.get_all_models())
    default_price_list_item = serializers.HyperlinkedRelatedField(
        view_name='defaultpricelistitem-detail',
        lookup_field='uuid',
        queryset=models.DefaultPriceListItem.objects.all().select_related('resource_content_type'))

    class Meta:
        model = models.PriceListItem
        fields = ('url', 'uuid', 'units', 'value', 'service', 'default_price_list_item')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'default_price_list_item': {'lookup_field': 'uuid'}
        }
        protected_fields = ('service', 'default_price_list_item')

    def create(self, validated_data):
        try:
            return super(PriceListItemSerializer, self).create(validated_data)
        except IntegrityError:
            raise serializers.ValidationError(_('Price list item for service already exists.'))


class DefaultPriceListItemSerializer(serializers.HyperlinkedModelSerializer):
    value = serializers.FloatField()

    class Meta:
        model = models.DefaultPriceListItem
        fields = ('url', 'uuid', 'key', 'item_type', 'value', 'resource_type')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class MergedPriceListItemSerializer(serializers.HyperlinkedModelSerializer):
    value = serializers.SerializerMethodField()
    units = serializers.SerializerMethodField()
    is_manually_input = serializers.SerializerMethodField()
    service_price_list_item_url = serializers.SerializerMethodField()

    class Meta:
        model = models.DefaultPriceListItem
        fields = ('url', 'uuid', 'key', 'item_type', 'units', 'value',
                  'resource_type', 'is_manually_input', 'service_price_list_item_url')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_value(self, obj):
        return getattr(obj, 'service_item', None) and float(obj.service_item[0].value) or float(obj.value)

    def get_units(self, obj):
        return getattr(obj, 'service_item', None) and obj.service_item[0].units or obj.units

    def get_is_manually_input(self, obj):
        return bool(getattr(obj, 'service_item', None))

    def get_service_price_list_item_url(self, obj):
        if not getattr(obj, 'service_item', None):
            return
        return reverse('pricelistitem-detail',
                       kwargs={'uuid': obj.service_item[0].uuid.hex},
                       request=self.context['request'])



from django.contrib import admin

from . import models


class QuestionInline(admin.TabularInline):
    model = models.Question
    fields = ('order', 'description', 'solution', 'category')


class ChecklistAdmin(admin.ModelAdmin):
    inlines = [QuestionInline]
    list_display = ('name', 'description')
    fields = ('name', 'description')


admin.site.register(models.Checklist, ChecklistAdmin)



import logging
import tempfile

import docker
from django.conf import settings
from docker.errors import DockerException
from rest_framework import serializers as rf_serializers

from . import serializers


logger = logging.getLogger(__name__)


def execute_script(image, command, src, **kwargs):
    with tempfile.NamedTemporaryFile(prefix='docker', mode="w+") as docker_script:
        docker_script.write(src)
        docker_script.flush()
        client = docker.DockerClient(**settings.WALDUR_MARKETPLACE_SCRIPT['DOCKER_CLIENT'])
        return client.containers.run(
            image=image,
            command=[command, 'script'],
            remove=True,
            working_dir="/work",
            volumes={
                docker_script.name: {
                    "bind": "/work/script",
                    "mode": "ro",
                },
            },
            **settings.WALDUR_MARKETPLACE_SCRIPT['DOCKER_RUN_OPTIONS'],
            **kwargs,
        )


class DockerExecutorMixin:
    hook_type = NotImplemented

    def send_request(self, user):
        options = self.order_item.offering.plugin_options

        serializer = serializers.OrderItemSerializer(instance=self.order_item)
        environment = {key.upper(): str(value) for key, value in serializer.data}
        if isinstance(options.get('environ'), dict):
            environment.update(options['environ'])

        language = options['language']
        image = settings.WALDUR_MARKETPLACE_SCRIPT['DOCKER_IMAGES'].get(language)
        logger.debug('About to execute marketplace script via Docker. '
                     'Hook type is %s. Order item ID is %s.',
                     self.hook_type, self.order_item.id)

        try:
            self.order_item.output = execute_script(
                image=image,
                command=language,
                src=options[self.hook_type],
                environment=environment
            )
            self.order_item.save(update_fields=['output'])
        except DockerException as exc:
            logger.exception('Unable to execute marketplace script via Docker. '
                             'Hook type is %s. Order item ID is %s.',
                             self.hook_type, self.order_item.id)
            raise rf_serializers.ValidationError(str(exc))
        logger.debug('Successfully executed marketplace script via Docker.'
                     'Hook type is %s. Order item ID is %s.',
                     self.hook_type, self.order_item.id)

    def validate_order_item(self, request):
        options = self.order_item.offering.plugin_options

        if self.hook_type not in options:
            raise rf_serializers.ValidationError('Script is not defined.')

        command = settings.WALDUR_MARKETPLACE_SCRIPT['DOCKER_IMAGES'].get(options['language'])
        if not command:
            raise rf_serializers.ValidationError('Docker image is not allowed.')



from django.contrib import admin
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _

from waldur_core.core.admin import ExecutorAdminAction
from waldur_core.structure import admin as structure_admin

from . import executors, models


class FlavorAdmin(structure_admin.BackendModelAdmin):
    list_display = ('name', 'settings', 'cores', 'ram')


class FloatingIPAdmin(structure_admin.BackendModelAdmin):
    list_display = ('address', 'settings', 'is_available')


class InternalIPAdmin(structure_admin.BackendModelAdmin):
    list_display = ('address', 'settings', 'is_available')


class VolumeAdmin(structure_admin.ResourceAdmin):

    class Pull(ExecutorAdminAction):
        executor = executors.VolumePullExecutor
        short_description = _('Pull')

        def validate(self, instance):
            if instance.state not in (models.Volume.States.OK, models.Volume.States.ERRED):
                raise ValidationError(_('Volume has to be in OK or ERRED state.'))

    pull = Pull()


class InstanceAdmin(structure_admin.VirtualMachineAdmin):
    actions = structure_admin.VirtualMachineAdmin.actions + ['pull']

    class Pull(ExecutorAdminAction):
        executor = executors.InstancePullExecutor
        short_description = _('Pull')

        def validate(self, instance):
            if instance.state not in (models.Instance.States.OK, models.Instance.States.ERRED):
                raise ValidationError(_('Instance has to be in OK or ERRED state.'))

    pull = Pull()


admin.site.register(models.RijkscloudService, structure_admin.ServiceAdmin)
admin.site.register(models.RijkscloudServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)
admin.site.register(models.Flavor, FlavorAdmin)
admin.site.register(models.FloatingIP, FloatingIPAdmin)
admin.site.register(models.InternalIP, InternalIPAdmin)
admin.site.register(models.Volume, VolumeAdmin)
admin.site.register(models.Instance, InstanceAdmin)



from django.core.exceptions import ObjectDoesNotExist
from rest_framework import serializers
from rest_framework.reverse import reverse

from waldur_mastermind.marketplace import processors, signals
from waldur_mastermind.packages import models as package_models
from waldur_mastermind.packages import views as package_views
from waldur_openstack.openstack import models as openstack_models
from waldur_openstack.openstack import views as openstack_views
from waldur_openstack.openstack_tenant import views as tenant_views

from . import utils


class TenantCreateProcessor(processors.CreateResourceProcessor):
    def get_serializer_class(self):
        return package_views.OpenStackPackageViewSet.create_serializer_class

    def get_viewset(self):
        return package_views.OpenStackPackageViewSet

    def get_post_data(self):
        order_item = self.order_item

        try:
            template = order_item.plan.scope
        except ObjectDoesNotExist:
            template = None
        except AttributeError:
            template = None

        if not isinstance(template, package_models.PackageTemplate):
            raise serializers.ValidationError('Plan has invalid scope. VPC package template is expected.')

        project = order_item.order.project

        project_url = reverse('project-detail', kwargs={'uuid': project.uuid.hex})
        spl_url = processors.get_spl_url(openstack_models.OpenStackServiceProjectLink, order_item)

        fields = (
            'name',
            'description',
            'user_username',
            'user_password',
            'subnet_cidr',
            'skip_connection_extnet',
            'availability_zone',
        )

        quotas = utils.map_limits_to_quotas(order_item.limits)

        return dict(
            project=project_url,
            service_project_link=spl_url,
            template=template.uuid.hex,
            quotas=quotas,
            **processors.copy_attributes(fields, order_item)
        )

    def get_scope_from_response(self, response):
        return package_models.OpenStackPackage.objects.get(uuid=response.data['uuid']).tenant


class TenantUpdateProcessor(processors.UpdateResourceProcessor):

    def get_serializer_class(self):
        return package_views.OpenStackPackageViewSet.change_serializer_class

    def get_view(self):
        return package_views.OpenStackPackageViewSet.as_view({'post': 'change'})

    def get_post_data(self):
        resource = self.get_resource()
        try:
            package = package_models.OpenStackPackage.objects.get(tenant=resource)
        except ObjectDoesNotExist:
            raise serializers.ValidationError('OpenStack package for tenant does not exist.')

        template = self.order_item.plan.scope

        return {
            'package': package.uuid.hex,
            'template': template.uuid.hex,
        }

    def update_limits_process(self, user):
        scope = self.order_item.resource.scope
        if not scope or not isinstance(scope, openstack_models.Tenant):
            signals.limit_update_failed.send(
                sender=self.order_item.resource.__class__,
                order_item=self.order_item,
                message='Limit updating is available only for tenants.'
            )
            return

        utils.update_limits(self.order_item)


class TenantDeleteProcessor(processors.DeleteResourceProcessor):
    viewset = openstack_views.TenantViewSet


class InstanceCreateProcessor(processors.BaseCreateResourceProcessor):
    viewset = tenant_views.InstanceViewSet

    fields = (
        'name',
        'description',
        'flavor',
        'image',
        'security_groups',
        'internal_ips_set',
        'floating_ips',
        'system_volume_size',
        'system_volume_type',
        'data_volume_size',
        'data_volume_type',
        'volumes',
        'ssh_public_key',
        'user_data',
        'availability_zone',
    )


class InstanceDeleteProcessor(processors.DeleteResourceProcessor):
    viewset = tenant_views.InstanceViewSet


class VolumeCreateProcessor(processors.BaseCreateResourceProcessor):
    viewset = tenant_views.VolumeViewSet

    fields = (
        'name',
        'description',
        'image',
        'size',
        'availability_zone',
        'type',
    )


class VolumeDeleteProcessor(processors.DeleteResourceProcessor):
    viewset = tenant_views.VolumeViewSet



from django.apps import AppConfig
from django.db.models import signals


class SlurmConfig(AppConfig):
    name = 'waldur_slurm'
    verbose_name = 'SLURM'
    service_name = 'SLURM'

    def ready(self):
        from waldur_core.quotas.fields import QuotaField, CounterQuotaField
        from waldur_core.structure import SupportedServices
        from waldur_core.structure import models as structure_models
        from waldur_core.structure import signals as structure_signals
        from waldur_freeipa import models as freeipa_models

        from .backend import SlurmBackend
        from . import handlers, models, utils

        SupportedServices.register_backend(SlurmBackend)

        signals.post_save.connect(
            handlers.process_user_creation,
            sender=freeipa_models.Profile,
            dispatch_uid='waldur_slurm.handlers.process_user_creation',
        )

        signals.pre_delete.connect(
            handlers.process_user_deletion,
            sender=freeipa_models.Profile,
            dispatch_uid='waldur_slurm.handlers.process_user_deletion',
        )

        structure_models_with_roles = (structure_models.Customer, structure_models.Project)
        for model in structure_models_with_roles:
            structure_signals.structure_role_granted.connect(
                handlers.process_role_granted,
                sender=model,
                dispatch_uid='waldur_slurm.handlers.process_role_granted.%s' % model.__class__,
            )

            structure_signals.structure_role_revoked.connect(
                handlers.process_role_revoked,
                sender=model,
                dispatch_uid='waldur_slurm.handlers.process_role_revoked.%s' % model.__class__,
            )

        for quota in utils.QUOTA_NAMES:
            structure_models.Customer.add_quota_field(
                name=quota,
                quota_field=QuotaField(is_backend=True)
            )

            structure_models.Project.add_quota_field(
                name=quota,
                quota_field=QuotaField(is_backend=True)
            )

        structure_models.Project.add_quota_field(
            name='nc_allocation_count',
            quota_field=CounterQuotaField(
                target_models=lambda: [models.Allocation],
                path_to_scope='service_project_link.project',
            )
        )

        structure_models.Customer.add_quota_field(
            name='nc_allocation_count',
            quota_field=CounterQuotaField(
                target_models=lambda: [models.Allocation],
                path_to_scope='service_project_link.project.customer',
            )
        )

        signals.post_save.connect(
            handlers.update_quotas_on_allocation_usage_update,
            sender=models.Allocation,
            dispatch_uid='waldur_slurm.handlers.update_quotas_on_allocation_usage_update',
        )



from django.apps import AppConfig
from django.db.models import signals


class FreeIPAConfig(AppConfig):
    name = 'waldur_freeipa'
    verbose_name = 'FreeIPA'

    def ready(self):
        from waldur_core.core import models as core_models
        from waldur_core.quotas.fields import QuotaField
        from waldur_core.quotas import models as quota_models
        from waldur_core.structure import models as structure_models
        from waldur_core.structure import signals as structure_signals

        from . import handlers, utils, models

        for model in (structure_models.Customer, structure_models.Project):
            signals.post_save.connect(
                handlers.schedule_sync,
                sender=model,
                dispatch_uid='waldur_freeipa.handlers.schedule_sync_on_%s_creation' % model.__class__,
            )

            signals.pre_delete.connect(
                handlers.schedule_sync,
                sender=model,
                dispatch_uid='waldur_freeipa.handlers.schedule_sync_on_%s_deletion' % model.__class__,
            )

            structure_signals.structure_role_granted.connect(
                handlers.schedule_sync,
                sender=model,
                dispatch_uid='waldur_freeipa.handlers.schedule_sync_on_%s_role_granted' % model.__class__,
            )

            structure_signals.structure_role_revoked.connect(
                handlers.schedule_sync,
                sender=model,
                dispatch_uid='waldur_freeipa.handlers.schedule_sync_on_%s_role_revoked' % model.__class__,
            )

        signals.post_save.connect(
            handlers.schedule_sync_on_quota_change,
            sender=quota_models.Quota,
            dispatch_uid='waldur_freeipa.handlers.schedule_sync_on_quota_save',
        )

        signals.post_save.connect(
            handlers.schedule_ssh_key_sync_when_key_is_created,
            sender=core_models.SshPublicKey,
            dispatch_uid='waldur_freeipa.handlers.schedule_ssh_key_sync_when_key_is_created',
        )

        signals.pre_delete.connect(
            handlers.schedule_ssh_key_sync_when_key_is_deleted,
            sender=core_models.SshPublicKey,
            dispatch_uid='waldur_freeipa.handlers.schedule_ssh_key_sync_when_key_is_deleted',
        )

        structure_models.Customer.add_quota_field(
            name=utils.QUOTA_NAME,
            quota_field=QuotaField()
        )

        structure_models.Project.add_quota_field(
            name=utils.QUOTA_NAME,
            quota_field=QuotaField()
        )

        signals.pre_save.connect(
            handlers.log_profile_event,
            sender=models.Profile,
            dispatch_uid='waldur_freeipa.handlers.log_profile_event',
        )

        signals.pre_delete.connect(
            handlers.log_profile_deleted,
            sender=models.Profile,
            dispatch_uid='waldur_freeipa.handlers.log_profile_deleted',
        )



from django.db import models
from django.utils.translation import ugettext_lazy as _
from libcloud.compute.drivers.ec2 import REGION_DETAILS

from waldur_core.core.fields import JSONField
from waldur_core.core.models import RuntimeStateMixin
from waldur_core.quotas.fields import CounterQuotaField
from waldur_core.quotas.models import QuotaModelMixin
from waldur_core.structure import models as structure_models
from waldur_geo_ip.utils import get_coordinates_by_ip


class AWSService(structure_models.Service):
    projects = models.ManyToManyField(
        structure_models.Project, related_name='aws_services', through='AWSServiceProjectLink')

    class Meta(structure_models.Service.Meta):
        verbose_name = _('AWS provider')
        verbose_name_plural = _('AWS providers')

    class Quotas(QuotaModelMixin.Quotas):
        instance_count = CounterQuotaField(
            target_models=lambda: [Instance],
            path_to_scope='service_project_link.service'
        )

        volume_count = CounterQuotaField(
            target_models=lambda: [Volume],
            path_to_scope='service_project_link.service'
        )

    @classmethod
    def get_url_name(cls):
        return 'aws'


class AWSServiceProjectLink(structure_models.CloudServiceProjectLink):
    service = models.ForeignKey(on_delete=models.CASCADE, to=AWSService)

    class Meta(structure_models.CloudServiceProjectLink.Meta):
        verbose_name = _('AWS provider project link')
        verbose_name_plural = _('AWS provider project links')

    @classmethod
    def get_url_name(cls):
        return 'aws-spl'


class Region(structure_models.GeneralServiceProperty):
    class Meta:
        ordering = ['name']

    @classmethod
    def get_url_name(cls):
        return 'aws-region'


class Image(structure_models.GeneralServiceProperty):
    class Meta:
        ordering = ['name']

    region = models.ForeignKey(on_delete=models.CASCADE, to=Region)

    def __str__(self):
        return '{0} | {1}'.format(self.name, self.region.name)

    @classmethod
    def get_url_name(cls):
        return 'aws-image'

    @classmethod
    def get_backend_fields(cls):
        return super(Image, cls).get_backend_fields() + ('region',)


class Size(structure_models.GeneralServiceProperty):
    class Meta:
        ordering = ['cores', 'ram']

    regions = models.ManyToManyField(Region)
    cores = models.PositiveSmallIntegerField(help_text=_('Number of cores in a VM'))
    ram = models.PositiveIntegerField(help_text=_('Memory size in MiB'))
    disk = models.PositiveIntegerField(help_text=_('Disk size in MiB'))
    price = models.DecimalField(_('Hourly price rate'), default=0, max_digits=11, decimal_places=5)

    @classmethod
    def get_url_name(cls):
        return 'aws-size'

    @classmethod
    def get_backend_fields(cls):
        return super(Size, cls).get_backend_fields() + ('cores', 'ram', 'disk', 'price', 'regions')


class Instance(structure_models.VirtualMachine):
    service_project_link = models.ForeignKey(
        AWSServiceProjectLink, related_name='instances', on_delete=models.PROTECT)

    region = models.ForeignKey(on_delete=models.CASCADE, to=Region)
    public_ips = JSONField(default=list, help_text=_('List of public IP addresses'), blank=True)
    private_ips = JSONField(default=list, help_text=_('List of private IP addresses'), blank=True)
    size_backend_id = models.CharField(max_length=150, blank=True)

    def increase_backend_quotas_usage(self, validate=True):
        spl = self.service_project_link
        spl.add_quota_usage(spl.Quotas.storage, self.disk, validate=validate)
        spl.add_quota_usage(spl.Quotas.ram, self.ram, validate=validate)
        spl.add_quota_usage(spl.Quotas.vcpu, self.cores, validate=validate)

    def decrease_backend_quotas_usage(self):
        self.service_project_link.add_quota_usage(self.service_project_link.Quotas.storage, -self.disk)
        self.service_project_link.add_quota_usage(self.service_project_link.Quotas.ram, -self.ram)
        self.service_project_link.add_quota_usage(self.service_project_link.Quotas.vcpu, -self.cores)

    @property
    def external_ips(self):
        return self.public_ips

    @property
    def internal_ips(self):
        return self.private_ips

    def detect_coordinates(self):
        if self.external_ips:
            return get_coordinates_by_ip(self.external_ips[0])
        region = self.region.backend_id
        endpoint = REGION_DETAILS[region]['endpoint']
        return get_coordinates_by_ip(endpoint)

    @classmethod
    def get_url_name(cls):
        return 'aws-instance'

    @classmethod
    def get_backend_fields(cls):
        return super(Instance, cls).get_backend_fields() + ('runtime_state',)

    @classmethod
    def get_online_state(cls):
        return 'running'

    @classmethod
    def get_offline_state(cls):
        return 'stopped'


class Volume(RuntimeStateMixin, structure_models.NewResource):
    service_project_link = models.ForeignKey(
        AWSServiceProjectLink, related_name='volumes', on_delete=models.PROTECT)

    VOLUME_TYPES = (
        ('gp2', _('General Purpose SSD')),
        ('io1', _('Provisioned IOPS SSD')),
        ('standard', _('Magnetic volumes'))
    )
    size = models.PositiveIntegerField(help_text=_('Size of volume in gigabytes'))
    region = models.ForeignKey(on_delete=models.CASCADE, to=Region)
    volume_type = models.CharField(max_length=8, choices=VOLUME_TYPES)
    device = models.CharField(max_length=128, blank=True, null=True)
    instance = models.ForeignKey(on_delete=models.CASCADE, to=Instance, blank=True, null=True)

    @classmethod
    def get_url_name(cls):
        return 'aws-volume'

    @classmethod
    def get_backend_fields(cls):
        return super(Volume, cls).get_backend_fields() + ('name', 'device', 'size', 'volume_type', 'runtime_state')



from waldur_core.core import WaldurExtension


class PIDExtension(WaldurExtension):

    class Settings:
        WALDUR_PID = {
            'DATACITE': {
                'REPOSITORY_ID': '',
                'PASSWORD': '',
                'PREFIX': '',
                'API_URL': 'https://example.com',
                'PUBLISHER': 'Waldur'
            }
        }

    @staticmethod
    def django_app():
        return 'waldur_pid'



from waldur_vmware import views as vmware_views
from waldur_mastermind.marketplace import processors


class VirtualMachineCreateProcessor(processors.BaseCreateResourceProcessor):
    viewset = vmware_views.VirtualMachineViewSet
    fields = (
        'name',
        'description',
        'guest_os',
        'cores_per_socket',
        'template',
        'cluster',
        'datastore',
    )

    def get_post_data(self):
        payload = super(VirtualMachineCreateProcessor, self).get_post_data()

        limits = self.order_item.limits
        if limits:
            if 'cpu' in limits:
                payload['cores'] = limits['cpu']
            if 'ram' in limits:
                payload['ram'] = limits['ram']
        return payload



from waldur_core.core import WaldurExtension


class OpenStackExtension(WaldurExtension):

    class Settings:
        # wiki: https://opennode.atlassian.net/wiki/display/WD/OpenStack+plugin+configuration
        WALDUR_OPENSTACK = {
            'DEFAULT_SECURITY_GROUPS': (
                {
                    'name': 'allow-all',
                    'description': 'Security group for any access',
                    'rules': (
                        {
                            'protocol': 'icmp',
                            'cidr': '0.0.0.0/0',
                            'icmp_type': -1,
                            'icmp_code': -1,
                        },
                        {
                            'protocol': 'tcp',
                            'cidr': '0.0.0.0/0',
                            'from_port': 1,
                            'to_port': 65535,
                        },
                    ),
                },
                {
                    'name': 'ssh',
                    'description': 'Security group for secure shell access',
                    'rules': (
                        {
                            'protocol': 'tcp',
                            'cidr': '0.0.0.0/0',
                            'from_port': 22,
                            'to_port': 22,
                        },
                    ),
                },
                {
                    'name': 'ping',
                    'description': 'Security group for ping',
                    'rules': (
                        {
                            'protocol': 'icmp',
                            'cidr': '0.0.0.0/0',
                            'icmp_type': -1,
                            'icmp_code': -1,
                        },
                    ),
                },
                {
                    'name': 'rdp',
                    'description': 'Security group for remove desktop access',
                    'rules': (
                        {
                            'protocol': 'tcp',
                            'cidr': '0.0.0.0/0',
                            'from_port': 3389,
                            'to_port': 3389,
                        },
                    ),
                },
                {
                    'name': 'web',
                    'description': 'Security group for http and https access',
                    'rules': (
                        {
                            'protocol': 'tcp',
                            'cidr': '0.0.0.0/0',
                            'from_port': 80,
                            'to_port': 80,
                        },
                        {
                            'protocol': 'tcp',
                            'cidr': '0.0.0.0/0',
                            'from_port': 443,
                            'to_port': 443,
                        },
                    ),
                },
            ),
            'SUBNET': {
                'ALLOCATION_POOL_START': '{first_octet}.{second_octet}.{third_octet}.10',
                'ALLOCATION_POOL_END': '{first_octet}.{second_octet}.{third_octet}.200',
            },
            'DEFAULT_BLACKLISTED_USERNAMES': ['admin', 'service'],
            # TODO: Delete these flags after migration to marketplace is completed
            # They are superseded by MANAGER_CAN_APPROVE_ORDER and ADMIN_CAN_APPROVE_ORDER
            # If this flag is true - manager can execute actions that will
            # change cost of the project: delete tenants, change their configuration
            'MANAGER_CAN_MANAGE_TENANTS': False,
            'ADMIN_CAN_MANAGE_TENANTS': False,
            'TENANT_CREDENTIALS_VISIBLE': True
        }

    @staticmethod
    def django_app():
        return 'waldur_openstack.openstack'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        from datetime import timedelta
        return {
            'openstack-tenant-pull-quotas': {
                'task': 'openstack.TenantPullQuotas',
                'schedule': timedelta(hours=24),
                'args': (),
            },
        }

    @staticmethod
    def get_cleanup_executor():
        from .executors import OpenStackCleanupExecutor
        return OpenStackCleanupExecutor

    @staticmethod
    def get_public_settings():
        return [
            'MANAGER_CAN_MANAGE_TENANTS',
            'TENANT_CREDENTIALS_VISIBLE',
        ]



from django.contrib import admin
from django.contrib.admin import options
from django.core.exceptions import ValidationError
from django.forms.models import BaseInlineFormSet
from django.utils.translation import ugettext_lazy as _

from waldur_core.core.admin import ExecutorAdminAction
from waldur_core.structure import admin as structure_admin
from waldur_vmware.utils import is_basic_mode

from . import executors, models


class DiskAdmin(structure_admin.ResourceAdmin):

    class Pull(ExecutorAdminAction):
        executor = executors.DiskPullExecutor
        short_description = _('Pull')

        def validate(self, instance):
            if instance.state not in (models.Disk.States.OK, models.Disk.States.ERRED):
                raise ValidationError(_('Disk has to be in OK or ERRED state.'))

    pull = Pull()


class VirtualMachineAdmin(structure_admin.ResourceAdmin):

    class Pull(ExecutorAdminAction):
        executor = executors.VirtualMachinePullExecutor
        short_description = _('Pull')

        def validate(self, instance):
            if instance.state not in (models.VirtualMachine.States.OK, models.VirtualMachine.States.ERRED):
                raise ValidationError(_('Virtual machine has to be in OK or ERRED state.'))

    pull = Pull()


class CustomerInlineFormset(BaseInlineFormSet):
    service_property_field = None

    def clean(self):
        """
        When basic mode is activated we should require one service property
        (network, cluster and folder) defined per customer
        per shared service setting.
        """
        super(CustomerInlineFormset, self).clean()
        if is_basic_mode():
            enabled_settings = {}
            for form in self.forms:
                cleaned_data = getattr(form, 'cleaned_data', None)

                # Skip empty form
                if not cleaned_data:
                    continue

                # Skip deleted form
                if cleaned_data.get('DELETE'):
                    continue

                # Ensure that the same service settings are not used multiple times
                service_settings = cleaned_data[self.service_property_field].settings
                if service_settings in enabled_settings:
                    raise ValidationError(_('There should be exactly one property '
                                            'assigned to the each service settings.'))
                else:
                    enabled_settings[service_settings] = True


class CustomerClusterInlineFormset(CustomerInlineFormset):
    service_property_field = 'cluster'


class CustomerClusterInline(options.TabularInline):
    model = models.CustomerCluster
    extra = 1
    verbose_name_plural = 'Customer VMware clusters'
    formset = CustomerClusterInlineFormset


class CustomerNetworkInlineFormset(CustomerInlineFormset):
    service_property_field = 'network'


class CustomerNetworkInline(options.TabularInline):
    model = models.CustomerNetwork
    extra = 1
    verbose_name_plural = 'Customer VMware networks for new VMs'
    formset = CustomerNetworkInlineFormset


class CustomerNetworkPairInline(options.TabularInline):
    model = models.CustomerNetworkPair
    extra = 1
    verbose_name_plural = 'Customer VMware networks for existing VMs'


class CustomerDatastoreInline(options.TabularInline):
    model = models.CustomerDatastore
    extra = 1
    verbose_name_plural = 'Customer VMware datastores'


class CustomerFolderInlineInlineFormset(CustomerInlineFormset):
    service_property_field = 'folder'


class CustomerFolderInline(options.TabularInline):
    model = models.CustomerFolder
    extra = 1
    verbose_name_plural = 'Customer VMware folders'
    formset = CustomerFolderInlineInlineFormset


admin.site.register(models.VMwareService, structure_admin.ServiceAdmin)
admin.site.register(models.VMwareServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)
admin.site.register(models.Disk, DiskAdmin)
admin.site.register(models.VirtualMachine, VirtualMachineAdmin)
admin.site.register(models.Template, structure_admin.ServicePropertyAdmin)
admin.site.register(models.Cluster, structure_admin.ServicePropertyAdmin)
admin.site.register(models.Datastore, structure_admin.ServicePropertyAdmin)
admin.site.register(models.Network, structure_admin.ServicePropertyAdmin)
admin.site.register(models.Folder, structure_admin.ServicePropertyAdmin)

structure_admin.CustomerAdmin.inlines += [CustomerClusterInline]
structure_admin.CustomerAdmin.inlines += [CustomerNetworkInline]
structure_admin.CustomerAdmin.inlines += [CustomerNetworkPairInline]
structure_admin.CustomerAdmin.inlines += [CustomerDatastoreInline]
structure_admin.CustomerAdmin.inlines += [CustomerFolderInline]



from django.contrib.auth import get_user_model
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_core.structure import models as structure_models
from waldur_core.users import models

User = get_user_model()


class InvitationSerializer(serializers.HyperlinkedModelSerializer):
    project = serializers.HyperlinkedRelatedField(
        view_name='project-detail',
        lookup_field='uuid',
        queryset=structure_models.Project.objects.all(),
        required=False,
        allow_null=True
    )
    project_name = serializers.ReadOnlyField(source='project.name')
    customer = serializers.HyperlinkedRelatedField(
        view_name='customer-detail',
        lookup_field='uuid',
        queryset=structure_models.Customer.objects.all(),
        required=False,
        allow_null=True
    )
    customer_name = serializers.ReadOnlyField(source='customer.name')

    expires = serializers.DateTimeField(source='get_expiration_time', read_only=True)

    class Meta:
        model = models.Invitation
        detail_fields = (
            'full_name', 'native_name',
            'tax_number', 'phone_number',
            'organization', 'job_title',
        )
        fields = ('url', 'uuid', 'link_template', 'email', 'civil_number',
                  'project', 'project_role', 'project_name',
                  'customer', 'customer_role', 'customer_name',
                  'state', 'error_message', 'created', 'expires') + detail_fields
        read_only_fields = ('url', 'uuid', 'state', 'error_message', 'created', 'expires')
        extra_kwargs = {
            'url': {
                'lookup_field': 'uuid',
                'view_name': 'user-invitation-detail',
            },
            'project_role': {
                'required': False,
                'allow_null': True,
            },
            'customer_role': {
                'required': False,
                'allow_null': True,
            },
        }

    def validate(self, attrs):
        link_template = attrs['link_template']
        if '{uuid}' not in link_template:
            raise serializers.ValidationError({'link_template': _("Link template must include '{uuid}' parameter.")})

        project = attrs.get('project')
        customer = attrs.get('customer')

        project_role = attrs.get('project_role', '')
        customer_role = attrs.get('customer_role', '')

        if customer and project:
            raise serializers.ValidationError(_('Cannot create invitation to project and customer simultaneously.'))
        elif not (customer or project):
            raise serializers.ValidationError(_('Customer or project must be provided.'))
        elif (customer and not customer_role) or (customer_role and not customer):
            raise serializers.ValidationError({'customer_role': _('Customer and its role must be provided.')})
        elif (project and not project_role) or (project_role and not project):
            raise serializers.ValidationError({'project_role': _('Project and its role must be provided.')})

        return attrs

    def create(self, validated_data):
        validated_data['created_by'] = self.context['request'].user
        project = validated_data.get('project')
        if project:
            validated_data['customer'] = project.customer
        return super(InvitationSerializer, self).create(validated_data)



from celery import chain

from django.conf import settings

from waldur_core.core import executors as core_executors
from waldur_core.core import tasks as core_tasks, utils as core_utils
from waldur_core.core.models import StateMixin

from . import tasks


class ClusterCreateExecutor(core_executors.BaseExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, user):
        _tasks = [core_tasks.BackendMethodTask().si(
            serialized_instance,
            'create_cluster',
            state_transition='begin_creating')]
        _tasks += cls.create_nodes(instance.node_set.all(), user)
        _tasks += [core_tasks.PollRuntimeStateTask().si(
            serialized_instance,
            backend_pull_method='check_cluster_creating',
            success_state=settings.WALDUR_RANCHER['ACTIVE_CLUSTER_STATE'],
            erred_state='error'
        )]
        return chain(*_tasks)

    @classmethod
    def create_nodes(cls, nodes, user):
        _tasks = []
        for node in nodes:
            serialized_instance = core_utils.serialize_instance(node)
            _tasks.append(tasks.CreateNodeTask().si(
                serialized_instance,
                user_id=user.id,
            ))
            _tasks.append(tasks.PollRuntimeStateNodeTask().si(serialized_instance))
        return _tasks


class ClusterDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, user):
        if instance.node_set.count():
            return tasks.DeleteClusterNodesTask().si(
                serialized_instance,
                user_id=user.id,
            )
        else:
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                'delete_cluster',
                state_transition='begin_deleting')


class ClusterUpdateExecutor(core_executors.UpdateExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id and {'name'} & set(kwargs['updated_fields']):
            return core_tasks.BackendMethodTask().si(
                serialized_instance,
                'update_cluster',
                state_transition='begin_updating')
        else:
            return core_tasks.StateTransitionTask().si(
                serialized_instance,
                state_transition='begin_updating'
            )


class NodeCreateExecutor(core_executors.CreateExecutor):
    @classmethod
    def get_task_signature(cls, instance, serialized_instance, user):
        return chain(
            tasks.CreateNodeTask().si(
                serialized_instance,
                user_id=user.id,
            ),
            tasks.PollRuntimeStateNodeTask().si(serialized_instance)
        )


class NodeDeleteExecutor(core_executors.BaseExecutor):
    @classmethod
    def get_task_signature(cls, instance, serialized_instance, user):
        return tasks.DeleteNodeTask().si(
            serialized_instance,
            user_id=user.id,
        )

    @classmethod
    def pre_apply(cls, instance, **kwargs):
        """
        We can start deleting a node even if it does not have the status OK or Erred,
        because a virtual machine could already be created.
        """
        instance.state = StateMixin.States.DELETION_SCHEDULED
        instance.save(update_fields=['state'])


class ClusterPullExecutor(core_executors.ActionExecutor):

    @classmethod
    def get_task_signature(cls, cluster, serialized_cluster, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_cluster, 'pull_cluster', state_transition='begin_updating')



import logging
from uuid import uuid4

from celery.task import Task as CeleryTask
from celery.worker.request import Request
from django.db import IntegrityError, models as django_models
from django.db.models import ObjectDoesNotExist
from django_fsm import TransitionNotAllowed

from waldur_core.core import models, utils
from waldur_core.core.exceptions import RuntimeStateException

logger = logging.getLogger(__name__)


class StateChangeError(RuntimeError):
    pass


class Task(CeleryTask):
    """ Base class for tasks that are run by executors.

    Provides standard way for input data deserialization.
    """

    @classmethod
    def get_description(cls, *args, **kwargs):
        """ Add additional information about task to celery logs.

            Receives same parameters as method "run".
        """
        raise NotImplementedError()

    def run(self, serialized_instance, *args, **kwargs):
        """ Deserialize input data and start backend operation execution """
        try:
            instance = utils.deserialize_instance(serialized_instance)
        except ObjectDoesNotExist:
            raise ObjectDoesNotExist('Cannot restore instance from serialized object %s. Probably it was deleted.' %
                                     serialized_instance)

        self.args = args
        self.kwargs = kwargs

        self.pre_execute(instance)
        result = self.execute(instance, *self.args, **self.kwargs)
        self.post_execute(instance)
        if result and isinstance(result, django_models.Model):
            result = utils.serialize_instance(result)
        return result

    def pre_execute(self, instance):
        pass

    def execute(self, instance, *args, **kwargs):
        """ Execute backend operation """
        raise NotImplementedError('%s should implement method `execute`' % self.__class__.__name__)

    def post_execute(self, instance):
        pass


class EmptyTask(CeleryTask):
    def run(self, *args, **kwargs):
        pass


class StateTransitionTask(Task):
    """ Execute instance state transition, changes instance action and action details if defined.

        It is impossible to change object action without state transition.
    """

    @classmethod
    def get_description(cls, instance, *args, **kwargs):
        transition_method = kwargs.get('state_transition')
        return 'Change state of object "%s" using method "%s".' % (instance, transition_method)

    def state_transition(self, instance, transition_method, action=None, action_details=None):
        instance_description = '%s instance `%s` (PK: %s)' % (instance.__class__.__name__, instance, instance.pk)
        old_state = instance.human_readable_state
        try:
            getattr(instance, transition_method)()
            if action is not None:
                instance.action = action
            if action_details is not None:
                instance.action_details = action_details
            instance.save()
        except IntegrityError:
            message = (
                'Could not change state of %s, using method `%s` due to concurrent update' %
                (instance_description, transition_method))
            raise StateChangeError(message)
        except TransitionNotAllowed:
            message = (
                'Could not change state of %s, using method `%s`. Current instance state: %s.' %
                (instance_description, transition_method, instance.human_readable_state))
            raise StateChangeError(message)
        else:
            logger.info('State of %s changed from %s to %s, with method `%s`',
                        instance_description, old_state, instance.human_readable_state, transition_method)

    def pre_execute(self, instance):
        state_transition = self.kwargs.pop('state_transition', None)
        action = self.kwargs.pop('action', None)
        action_details = self.kwargs.pop('action_details', None)
        if state_transition is not None:
            self.state_transition(instance, state_transition, action, action_details)
        super(StateTransitionTask, self).pre_execute(instance)

    # Empty execute method allows to use StateTransitionTask as standalone task
    def execute(self, instance, *args, **kwargs):
        return instance


class RuntimeStateChangeTask(Task):
    """ Allows to change runtime state of instance before and after execution.

    Define kwargs:
     - runtime_state - to change instance runtime state during execution.
     - success_runtime_state - to change instance runtime state after success tasks execution.
    """

    @classmethod
    def get_description(cls, instance, *args, **kwargs):
        runtime_state = kwargs.get('runtime_state')
        return 'Change runtime state of object "%s" to "%s".' % (instance, runtime_state)

    def update_runtime_state(self, instance, runtime_state):
        instance.runtime_state = runtime_state
        instance.save(update_fields=['runtime_state'])

    def pre_execute(self, instance):
        self.runtime_state = self.kwargs.pop('runtime_state', None)
        self.success_runtime_state = self.kwargs.pop('success_runtime_state', None)

        if self.runtime_state is not None:
            self.update_runtime_state(instance, self.runtime_state)
        super(RuntimeStateChangeTask, self).pre_execute(instance)

    def post_execute(self, instance, *args, **kwargs):
        if self.success_runtime_state is not None:
            self.update_runtime_state(instance, self.success_runtime_state)
        super(RuntimeStateChangeTask, self).post_execute(instance)

    # Empty execute method allows to use RuntimeStateChangeTask as standalone task
    def execute(self, instance, *args, **kwargs):
        return instance


class BackendMethodTask(RuntimeStateChangeTask, StateTransitionTask):
    """ Execute method of instance backend """

    @classmethod
    def get_description(cls, instance, backend_method, *args, **kwargs):
        actions = ['Run backend method "%s" for instance "%s".' % (backend_method, instance)]
        if 'state_transition' in kwargs:
            actions.append(StateTransitionTask.get_description(instance, backend_method, *args, **kwargs))
        if 'runtime_state' in kwargs:
            actions.append(RuntimeStateChangeTask.get_description(instance, backend_method, *args, **kwargs))
        return ' '.join(actions)

    def get_backend(self, instance):
        return instance.get_backend()

    def execute(self, instance, backend_method, *args, **kwargs):
        backend = self.get_backend(instance)
        return getattr(backend, backend_method)(instance, *args, **kwargs)


class IndependentBackendMethodTask(BackendMethodTask):
    """ Execute instance backend method that does not receive instance as argument """

    def execute(self, instance, backend_method, *args, **kwargs):
        backend = self.get_backend(instance)
        return getattr(backend, backend_method)(*args, **kwargs)


class DeletionTask(Task):
    """ Delete instance """

    @classmethod
    def get_description(cls, *args, **kwargs):
        instance = args[0]
        return 'Delete instance "%s".' % instance

    def execute(self, instance):
        instance_description = '%s instance `%s` (PK: %s)' % (instance.__class__.__name__, instance, instance.pk)
        instance.delete()
        logger.info('%s was successfully deleted', instance_description)


class ErrorMessageTask(Task):
    """ Store error in error_message field.

    This task should not be called as immutable, because it expects result_uuid
    as input argument.
    """

    @classmethod
    def get_description(cls, result_id, instance, *args, **kwargs):
        return 'Add error message to instance "%s".' % instance

    def run(self, result_id, serialized_instance, *args, **kwargs):
        self.result = self.AsyncResult(result_id)
        return super(ErrorMessageTask, self).run(serialized_instance, *args, **kwargs)

    def save_error_message(self, instance):
        if isinstance(instance, models.ErrorMessageMixin):
            instance.error_message = self.result.result
            instance.save(update_fields=['error_message'])
            # log exception if instance is not already ERRED.
            if instance.state != models.StateMixin.States.ERRED:
                message = 'Instance: %s.\n' % utils.serialize_instance(instance)
                message += 'Error: %s.\n' % self.result.result
                message += self.result.traceback
                logger.exception(message)

    def execute(self, instance):
        self.save_error_message(instance)


class ErrorStateTransitionTask(ErrorMessageTask, StateTransitionTask):
    """ Set instance as erred and save error message.

    This task should not be called as immutable, because it expects result_uuid
    as input argument.
    """

    @classmethod
    def get_description(cls, result_id, instance, *args, **kwargs):
        return 'Add error message and set erred instance "%s".' % instance

    def execute(self, instance):
        self.save_error_message(instance)
        self.state_transition(instance, 'set_erred', action='', action_details={})


class RecoverTask(StateTransitionTask):
    """ Change instance state from ERRED to OK and clear error_message """

    @classmethod
    def get_description(cls, instance, *args, **kwargs):
        return 'Recover instance "%s".' % instance

    def execute(self, instance):
        self.state_transition(instance, 'recover')
        instance.error_message = ''
        instance.save(update_fields=['error_message'])


class PreApplyExecutorTask(Task):
    """ Run executor as a task """

    @classmethod
    def get_description(cls, executor, instance, *args, **kwargs):
        return 'Run executor "%s" for instance "%s".' % (executor, instance)

    def run(self, serialized_executor, serialized_instance, *args, **kwargs):
        self.executor = utils.deserialize_class(serialized_executor)
        return super(PreApplyExecutorTask, self).run(serialized_instance, *args, **kwargs)

    def execute(self, instance, **kwargs):
        self.executor.pre_apply(instance, **kwargs)


class BackgroundTask(CeleryTask):
    """ Task that is run in background via celerybeat.

        Background task features:
         - background task does not start if previous task with the same name
           and input parameters is not completed yet;
         - all background tasks are scheduled in separate queue "background";
         - by default we do not log background tasks in celery logs. So tasks
           should log themselves explicitly and make sure that they will not
           spam error messages.

        Implement "is_equal" method to define what tasks are equal and should
        be executed simultaneously.
    """
    is_background = True

    def is_equal(self, other_task, *args, **kwargs):
        """ Return True if task do the same operation as other_task.

            Note! Other task is represented as serialized celery task - dictionary.
        """
        raise NotImplementedError('BackgroundTask should implement "is_equal" method to avoid queue overload.')

    def is_previous_task_processing(self, *args, **kwargs):
        """ Return True if exist task that is equal to current and is uncompleted """
        app = self._get_app()
        inspect = app.control.inspect()
        active = inspect.active() or {}
        scheduled = inspect.scheduled() or {}
        reserved = inspect.reserved() or {}
        uncompleted = sum(list(active.values()) + list(scheduled.values()) + list(reserved.values()), [])
        return any(self.is_equal(task, *args, **kwargs) for task in uncompleted)

    def apply_async(self, args=None, kwargs=None, **options):
        """ Do not run background task if previous task is uncompleted """
        if self.is_previous_task_processing(*args, **kwargs):
            message = 'Background task %s was not scheduled, because its predecessor is not completed yet.' % self.name
            logger.info(message)
            # It is expected by Celery that apply_async return AsyncResult, otherwise celerybeat dies
            return self.AsyncResult(options.get('task_id') or str(uuid4()))
        return super(BackgroundTask, self).apply_async(args=args, kwargs=kwargs, **options)


def log_celery_task(request):
    """ Add description to celery log output """
    task = request.task
    description = None
    if isinstance(task, Task):
        try:
            args, kwargs, embed = request._payload
            description = task.get_description(*args, **kwargs)
        except NotImplementedError:
            pass
        except Exception as e:
            # Logging should never break workflow.
            logger.exception('Cannot get description for task %s. Error: %s' % (task.__class__.__name__, e))

    return '{0.name}[{0.id}]{1}{2}{3}'.format(
        request,
        ' {0}'.format(description) if description else '',
        ' eta:[{0}]'.format(request.eta) if request.eta else '',
        ' expires:[{0}]'.format(request.expires) if request.expires else '',
    )


# XXX: drop the hack and use shadow name in celery 4.0
Request.__str__ = log_celery_task


class PollRuntimeStateTask(Task):
    max_retries = 1200
    default_retry_delay = 5

    @classmethod
    def get_description(cls, instance, backend_pull_method, *args, **kwargs):
        return 'Poll instance "%s" with method "%s"' % (instance, backend_pull_method)

    def get_backend(self, instance):
        return instance.get_backend()

    def execute(self, instance, backend_pull_method, success_state, erred_state, deleted_state=None):
        backend = self.get_backend(instance)
        getattr(backend, backend_pull_method)(instance)
        instance.refresh_from_db()
        if instance.runtime_state not in (success_state, erred_state, deleted_state):
            self.retry()
        elif instance.runtime_state == erred_state:
            raise RuntimeStateException(
                '%s (PK: %s) runtime state become erred: %s' % (
                    instance.__class__.__name__, instance.pk, erred_state))
        return instance


class PollStateTask(Task):
    max_retries = 1200
    default_retry_delay = 5

    def execute(self, instance, *args, **kwargs):
        if instance.state not in (models.StateMixin.States.OK, models.StateMixin.States.ERRED):
            self.retry()


class PollBackendCheckTask(Task):
    max_retries = 600
    default_retry_delay = 5

    @classmethod
    def get_description(cls, instance, backend_check_method, *args, **kwargs):
        return 'Check instance "%s" with method "%s"' % (instance, backend_check_method)

    def get_backend(self, instance):
        return instance.get_backend()

    def execute(self, instance, backend_check_method):
        # backend_check_method should return True if object does not exist at backend
        backend = self.get_backend(instance)
        if not getattr(backend, backend_check_method)(instance):
            self.retry()
        return instance


class ExtensionTaskMixin(CeleryTask):
    """
    This mixin allows to skip task scheduling if extension is disabled.
    Subclasses should implement "is_extension_disabled" method which returns boolean value.
    """
    def is_extension_disabled(self):
        return False

    def apply_async(self, args=None, kwargs=None, **options):
        if self.is_extension_disabled():
            message = 'Task %s is not scheduled, because its extension is disabled.' % self.name
            logger.info(message)
            return self.AsyncResult(options.get('task_id') or str(uuid4()))
        return super(ExtensionTaskMixin, self).apply_async(args=args, kwargs=kwargs, **options)



from django.dispatch import Signal

limit_update_succeeded = Signal(providing_args=['order_item'])
limit_update_failed = Signal(providing_args=['order_item', 'error_message'])
resource_creation_succeeded = Signal(providing_args=['instance'])
resource_update_succeeded = Signal(providing_args=['instance'])
resource_deletion_succeeded = Signal(providing_args=['instance'])



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-07-25 07:28
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_vmware', '0017_customer_network_pair'),
    ]

    operations = [
        migrations.AddField(
            model_name='port',
            name='runtime_state',
            field=models.CharField(blank=True, max_length=150, verbose_name='runtime state'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-03-20 13:03
from django.db import migrations, models
import django.db.models.deletion
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0007_customer_blocked'),
    ]

    operations = [
        migrations.CreateModel(
            name='Division',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='structure.Division')),
            ],
            options={
                'ordering': ('name',),
                'verbose_name': 'division',
            },
        ),
        migrations.CreateModel(
            name='DivisionType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
            ],
            options={
                'ordering': ('name',),
                'verbose_name': 'division type',
            },
        ),
        migrations.AddField(
            model_name='division',
            name='type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.DivisionType'),
        ),
        migrations.AddField(
            model_name='customer',
            name='division',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='structure.Division'),
        ),
    ]



from waldur_rancher import views as rancher_views
from waldur_mastermind.marketplace import processors


class RancherCreateProcessor(processors.BaseCreateResourceProcessor):
    viewset = rancher_views.ClusterViewSet
    fields = (
        'name',
        'description',
        'nodes',
        'tenant_settings',
    )


class RancherDeleteProcessor(processors.DeleteResourceProcessor):
    viewset = rancher_views.ClusterViewSet



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-22 09:32
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0008_customer_division'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='is_removed',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterModelOptions(
            name='project',
            options={'base_manager_name': 'objects'},
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-07-10 12:25
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_vmware', '0011_add_datastore'),
    ]

    operations = [
        migrations.AddField(
            model_name='template',
            name='disk',
            field=models.PositiveIntegerField(default=0, help_text='Disk size in MiB'),
        ),
    ]



from zipfile import is_zipfile, ZipFile

from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers, exceptions

from waldur_ansible.common import serializers as common_serializers
from waldur_core.core import models as core_models
from waldur_core.core import serializers as core_serializers
from waldur_core.core import utils as core_utils
from waldur_core.media.serializers import ProtectedMediaSerializerMixin, ProtectedFileField
from waldur_core.structure import permissions as structure
from waldur_core.structure import serializers as structure_serializers
from waldur_openstack.openstack_tenant import models as openstack_models

from . import models as playbook_jobs_models


class PlaybookParameterSerializer(serializers.ModelSerializer):
    name = serializers.RegexField('^[\w]+$')

    class Meta:
        model = playbook_jobs_models.PlaybookParameter
        fields = ('name', 'description', 'required', 'default')


class PlaybookSerializer(ProtectedMediaSerializerMixin,
                         core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):
    archive = serializers.FileField(write_only=True)
    parameters = PlaybookParameterSerializer(many=True, required=False)

    class Meta:
        model = playbook_jobs_models.Playbook
        fields = ('url', 'uuid', 'name', 'description', 'archive', 'entrypoint', 'parameters', 'image')
        protected_fields = ('entrypoint', 'parameters', 'archive')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def validate_archive(self, value):
        if not is_zipfile(value):
            raise serializers.ValidationError(_('ZIP file must be uploaded.'))
        elif not value.name.endswith('.zip'):
            raise serializers.ValidationError(_("File must have '.zip' extension."))

        zip_file = ZipFile(value)
        invalid_file = zip_file.testzip()
        if invalid_file is not None:
            raise serializers.ValidationError(
                _('File {filename} in archive {archive_name} has an invalid type.'.format(
                    filename=invalid_file, archive_name=zip_file.filename)))

        return value

    def validate(self, attrs):
        if self.instance:
            return attrs

        zip_file = ZipFile(attrs['archive'])
        entrypoint = attrs['entrypoint']
        if entrypoint not in zip_file.namelist():
            raise serializers.ValidationError(
                _('Failed to find entrypoint {entrypoint} in archive {archive_name}.'.format(
                    entrypoint=entrypoint, archive_name=zip_file.filename)))

        return attrs

    @transaction.atomic
    def create(self, validated_data):
        parameters_data = validated_data.pop('parameters', [])
        archive = validated_data.pop('archive')
        validated_data['workspace'] = playbook_jobs_models.Playbook.generate_workspace_path()

        zip_file = ZipFile(archive)
        zip_file.extractall(validated_data['workspace'])
        zip_file.close()

        playbook = playbook_jobs_models.Playbook.objects.create(**validated_data)
        for parameter_data in parameters_data:
            playbook_jobs_models.PlaybookParameter.objects.create(playbook=playbook, **parameter_data)

        return playbook


class JobSerializer(common_serializers.BaseApplicationSerializer,
                    structure_serializers.PermissionFieldFilteringMixin):
    service_project_link = serializers.HyperlinkedRelatedField(
        lookup_field='pk',
        view_name='openstacktenant-spl-detail',
        queryset=openstack_models.OpenStackTenantServiceProjectLink.objects.all(),
    )
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        lookup_field='uuid',
        view_name='openstacktenant-detail',
        read_only=True,
    )
    service_name = serializers.ReadOnlyField(source='service_project_link.service.settings.name')
    service_uuid = serializers.ReadOnlyField(source='service_project_link.service.uuid')
    ssh_public_key = serializers.HyperlinkedRelatedField(
        lookup_field='uuid',
        view_name='sshpublickey-detail',
        queryset=core_models.SshPublicKey.objects.all(),
        required=True,
    )
    ssh_public_key_name = serializers.ReadOnlyField(source='ssh_public_key.name')
    ssh_public_key_uuid = serializers.ReadOnlyField(source='ssh_public_key.uuid')
    project = serializers.HyperlinkedRelatedField(
        source='service_project_link.project',
        lookup_field='uuid',
        view_name='project-detail',
        read_only=True,
    )
    project_name = serializers.ReadOnlyField(source='service_project_link.project.name')
    project_uuid = serializers.ReadOnlyField(source='service_project_link.project.uuid')
    playbook = serializers.HyperlinkedRelatedField(
        lookup_field='uuid',
        view_name=core_utils.get_detail_view_name(playbook_jobs_models.Playbook),
        queryset=playbook_jobs_models.Playbook.objects.all(),
    )
    playbook_name = serializers.ReadOnlyField(source='playbook.name')
    playbook_uuid = serializers.ReadOnlyField(source='playbook.uuid')
    playbook_image = ProtectedFileField(source='playbook.image', read_only=True)
    playbook_description = serializers.ReadOnlyField(source='playbook.description')
    arguments = serializers.JSONField(required=False)
    state = serializers.SerializerMethodField()
    tag = serializers.SerializerMethodField()
    type = serializers.SerializerMethodField()

    class Meta:
        model = playbook_jobs_models.Job
        fields = ('url', 'uuid', 'name', 'description',
                  'service_project_link', 'service', 'service_name', 'service_uuid',
                  'ssh_public_key', 'ssh_public_key_name', 'ssh_public_key_uuid',
                  'project', 'project_name', 'project_uuid',
                  'playbook', 'playbook_name', 'playbook_uuid',
                  'playbook_image', 'playbook_description',
                  'arguments', 'state', 'output', 'created', 'modified', 'tag', 'type')
        read_only_fields = ('output', 'created', 'modified', 'type')
        protected_fields = ('service_project_link', 'ssh_public_key', 'playbook', 'arguments')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_type(self, obj):
        return 'playbook_job'

    def get_filtered_field_names(self):
        return 'project', 'service_project_link', 'ssh_public_key'

    def get_state(self, obj):
        return obj.get_state_display()

    def get_tag(self, obj):
        return obj.get_tag()

    def check_project(self, attrs):
        if self.instance:
            project = self.instance.service_project_link.project
        else:
            project = attrs['service_project_link'].project
        if not structure._has_admin_access(self.context['request'].user, project):
            raise exceptions.PermissionDenied()

    def check_arguments(self, attrs):
        playbook = self.instance.playbook if self.instance else attrs['playbook']
        arguments = attrs.get('arguments', {})
        parameter_names = playbook.parameters.all().values_list('name', flat=True)
        for argument in arguments.keys():
            if argument not in parameter_names and argument != 'project_uuid':
                raise serializers.ValidationError(_('Argument %s is not listed in playbook parameters.' % argument))

        if playbook.parameters.exclude(name__in=arguments.keys()).filter(required=True, default__exact='').exists():
            raise serializers.ValidationError(_('Not all required playbook parameters were specified.'))

        unfilled_parameters = playbook.parameters.exclude(name__in=arguments.keys())
        for parameter in unfilled_parameters:
            if parameter.default:
                arguments[parameter.name] = parameter.default

    def check_subnet(self, attrs):
        if not self.instance:
            settings = attrs['service_project_link'].service.settings
            if not openstack_models.SubNet.objects.filter(settings=settings).exists():
                raise serializers.ValidationError(_('Selected OpenStack provider does not have any subnet yet.'))
            else:
                attrs['subnet'] = openstack_models.SubNet.objects.filter(settings=settings).first()

    def validate(self, attrs):
        if not self.instance:
            attrs['user'] = self.context['request'].user

        self.check_project(attrs)
        self.check_arguments(attrs)
        self.check_subnet(attrs)
        return attrs



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-09-09 13:00
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import django_fsm
import model_utils.fields
import waldur_core.core.shims
import waldur_core.core.fields
import waldur_core.core.models
import waldur_core.core.validators
import waldur_core.logging.loggers


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0001_squashed_0054'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('taggit', '0002_auto_20150616_2121'),
        ('quotas', '0001_squashed_0004'),
    ]

    operations = [
        migrations.CreateModel(
            name='Flavor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('cores', models.PositiveSmallIntegerField(help_text='Number of cores in a VM')),
                ('ram', models.PositiveIntegerField(help_text='Memory size in MiB')),
                ('disk', models.PositiveIntegerField(help_text='Root disk size in MiB')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='FloatingIP',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('address', models.GenericIPAddressField(protocol='IPv4')),
                ('status', models.CharField(max_length=30)),
                ('backend_id', models.CharField(max_length=255)),
                ('backend_network_id', models.CharField(editable=False, max_length=255)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Floating IP',
                'verbose_name_plural': 'Floating IPs',
            },
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(db_index=True, max_length=255)),
                ('min_disk', models.PositiveIntegerField(default=0, help_text='Minimum disk size in MiB')),
                ('min_ram', models.PositiveIntegerField(default=0, help_text='Minimum memory size in MiB')),
                ('settings', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='OpenStackService',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('available_for_all', models.BooleanField(default=False, help_text='Service will be automatically added to all customers projects if it is available for all')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer', verbose_name='organization')),
            ],
            options={
                'verbose_name': 'OpenStack provider',
                'verbose_name_plural': 'OpenStack providers',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='OpenStackServiceProjectLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
                ('service', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='openstack.OpenStackService')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'OpenStack provider project link',
                'verbose_name_plural': 'OpenStack provider project links',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SecurityGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=128)),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='security_groups', to='openstack.OpenStackServiceProjectLink')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SecurityGroupRule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('protocol', models.CharField(blank=True, choices=[('tcp', 'tcp'), ('udp', 'udp'), ('icmp', 'icmp')], max_length=4)),
                ('from_port', models.IntegerField(null=True, validators=[django.core.validators.MaxValueValidator(65535)])),
                ('to_port', models.IntegerField(null=True, validators=[django.core.validators.MaxValueValidator(65535)])),
                ('cidr', models.CharField(blank=True, max_length=32)),
                ('backend_id', models.CharField(blank=True, max_length=128)),
                ('security_group', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='rules', to='openstack.SecurityGroup')),
            ],
        ),
        migrations.CreateModel(
            name='Tenant',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('start_time', models.DateTimeField(blank=True, null=True)),
                ('internal_network_id', models.CharField(blank=True, max_length=64)),
                ('external_network_id', models.CharField(blank=True, max_length=64)),
                ('availability_zone', models.CharField(blank=True, help_text='Optional availability group. Will be used for all instances provisioned in this tenant', max_length=100)),
                ('user_username', models.CharField(blank=True, max_length=50)),
                ('user_password', models.CharField(blank=True, max_length=50)),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='tenants', to='openstack.OpenStackServiceProjectLink')),
                ('tags', waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
                ('extra_configuration', waldur_core.core.fields.JSONField(default={}, help_text='Configuration details that are not represented on backend.')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='openstackservice',
            name='projects',
            field=models.ManyToManyField('structure.Project', related_name='openstack_services', through='openstack.OpenStackServiceProjectLink'),
        ),
        migrations.AddField(
            model_name='openstackservice',
            name='settings',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='service_project_link',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='floating_ips', to='openstack.OpenStackServiceProjectLink'),
        ),
        migrations.AlterUniqueTogether(
            name='openstackserviceprojectlink',
            unique_together=set([('service', 'project')]),
        ),
        migrations.RemoveField(
            model_name='openstackservice',
            name='name',
        ),
        migrations.AlterUniqueTogether(
            name='openstackservice',
            unique_together=set([('customer', 'settings')]),
        ),
        migrations.AlterUniqueTogether(
            name='image',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='flavor',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AddField(
            model_name='securitygroup',
            name='tenant',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='security_groups', to='openstack.Tenant'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='tenant',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='floating_ips', to='openstack.Tenant'),
        ),
        migrations.CreateModel(
            name='Network',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('runtime_state', models.CharField(blank=True, max_length=150, verbose_name='runtime state')),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('start_time', models.DateTimeField(blank=True, null=True)),
                ('is_external', models.BooleanField(default=False)),
                ('type', models.CharField(blank=True, max_length=50)),
                ('segmentation_id', models.IntegerField(null=True)),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='networks', to='openstack.OpenStackServiceProjectLink')),
                ('tags', waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
                ('tenant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='networks', to='openstack.Tenant')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='SubNet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('state', django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5)),
                ('backend_id', models.CharField(blank=True, max_length=255)),
                ('cidr', models.CharField(blank=True, max_length=32)),
                ('gateway_ip', models.GenericIPAddressField(null=True, protocol='IPv4')),
                ('allocation_pools', waldur_core.core.fields.JSONField(default=dict)),
                ('ip_version', models.SmallIntegerField(default=4)),
                ('enable_dhcp', models.BooleanField(default=True)),
                ('network', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subnets', to='openstack.Network')),
                ('service_project_link', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='subnets', to='openstack.OpenStackServiceProjectLink')),
                ('tags', waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags')),
                ('dns_nameservers', waldur_core.core.fields.JSONField(default=list, help_text='List of DNS name servers associated with the subnet.')),
                ('disable_gateway', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Subnet',
                'verbose_name_plural': 'Subnets',
            },
            bases=(waldur_core.core.models.DescendantMixin, waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='created',
            field=model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='description',
            field=models.CharField(blank=True, max_length=500, verbose_name='description'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='error_message',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='modified',
            field=model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='name',
            field=models.CharField(default='', max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='floatingip',
            name='runtime_state',
            field=models.CharField(blank=True, max_length=150, verbose_name='runtime state'),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='state',
            field=django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5),
        ),
        migrations.AddField(
            model_name='floatingip',
            name='tags',
            field=waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AlterField(
            model_name='floatingip',
            name='address',
            field=models.GenericIPAddressField(blank=True, null=True, protocol='IPv4'),
        ),
        migrations.AlterField(
            model_name='floatingip',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.RemoveField(
            model_name='floatingip',
            name='status',
        ),
        migrations.AddField(
            model_name='securitygroup',
            name='created',
            field=model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created'),
        ),
        migrations.AddField(
            model_name='securitygroup',
            name='modified',
            field=model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified'),
        ),
        migrations.AddField(
            model_name='securitygroup',
            name='tags',
            field=waldur_core.core.shims.TaggableManager(related_name='+', blank=True, help_text='A comma-separated list of tags.', through='taggit.TaggedItem', to='taggit.Tag', verbose_name='Tags'),
        ),
        migrations.AlterField(
            model_name='securitygroup',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.RemoveField(
            model_name='network',
            name='start_time',
        ),
        migrations.RemoveField(
            model_name='tenant',
            name='start_time',
        ),
        migrations.CreateModel(
            name='CustomerOpenStack',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('external_network_id', models.CharField(max_length=255, verbose_name='OpenStack external network ID')),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Customer')),
                ('settings', models.ForeignKey(limit_choices_to={'shared': True, 'type': 'OpenStack'}, on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings')),
            ],
            options={
                'verbose_name': 'Organization OpenStack settings',
                'verbose_name_plural': 'Organization OpenStack settings',
            },
        ),
        migrations.AlterUniqueTogether(
            name='customeropenstack',
            unique_together=set([('settings', 'customer')]),
        ),
        migrations.AlterField(
            model_name='tenant',
            name='backend_id',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='tenant',
            name='extra_configuration',
            field=waldur_core.core.fields.JSONField(default=dict, help_text='Configuration details that are not represented on backend.'),
        ),
        migrations.AddField(
            model_name='tenant',
            name='default_volume_type_name',
            field=models.CharField(blank=True, help_text='Volume type name to use when creating volumes.', max_length=100),
        ),
        migrations.AlterUniqueTogether(
            name='tenant',
            unique_together=set([('service_project_link', 'backend_id')]),
        ),
        migrations.AlterField(
            model_name='floatingip',
            name='address',
            field=models.GenericIPAddressField(blank=True, default=None, null=True, protocol='IPv4'),
        ),
        migrations.AlterUniqueTogether(
            name='floatingip',
            unique_together=set([('tenant', 'address')]),
        ),
    ]



from django.apps import AppConfig
from django.contrib.auth import get_user_model
from django.db.models import signals
from django_fsm import signals as fsm_signals

from waldur_core.core.monkeypatch import monkey_patch_fields


class CoreConfig(AppConfig):
    name = 'waldur_core.core'
    verbose_name = 'Core'

    def ready(self):
        from waldur_core.core import handlers
        from waldur_core.core.models import StateMixin
        from rest_framework.authtoken.models import Token

        User = get_user_model()
        SshPublicKey = self.get_model('SshPublicKey')

        signals.pre_save.connect(
            handlers.preserve_fields_before_update,
            sender=User,
            dispatch_uid='waldur_core.core.handlers.preserve_fields_before_update',
        )

        signals.post_save.connect(
            handlers.create_auth_token,
            sender=User,
            dispatch_uid='waldur_core.core.handlers.create_auth_token',
        )

        signals.post_save.connect(
            handlers.log_user_save,
            sender=User,
            dispatch_uid='waldur_core.core.handlers.log_user_save',
        )

        signals.post_save.connect(
            handlers.set_default_token_lifetime,
            sender=User,
            dispatch_uid='waldur_core.core.handlers.set_default_token_lifetime',
        )

        signals.post_delete.connect(
            handlers.log_user_delete,
            sender=User,
            dispatch_uid='waldur_core.core.handlers.log_user_delete',
        )

        signals.post_save.connect(
            handlers.log_ssh_key_save,
            sender=SshPublicKey,
            dispatch_uid='waldur_core.core.handlers.log_ssh_key_save',
        )

        signals.post_delete.connect(
            handlers.log_ssh_key_delete,
            sender=SshPublicKey,
            dispatch_uid='waldur_core.core.handlers.log_ssh_key_delete',
        )

        signals.post_save.connect(
            handlers.log_token_create,
            sender=Token,
            dispatch_uid='waldur_core.core.handlers.log_token_create',
        )

        for index, model in enumerate(StateMixin.get_all_models()):
            fsm_signals.post_transition.connect(
                handlers.delete_error_message,
                sender=model,
                dispatch_uid='waldur_core.core.handlers.delete_error_message_%s_%s' % (model.__name__, index),
            )

        # Database fields should be patched only after database models are initialized
        monkey_patch_fields()



from django.db import models

from model_utils.models import TimeStampedModel

from waldur_core.core import models as core_models
from waldur_core.structure.models import Project, StructureModel
from waldur_mastermind.marketplace.models import Category


class Checklist(core_models.UuidMixin,
                core_models.NameMixin,
                core_models.DescribableMixin,
                TimeStampedModel):

    def __str__(self):
        return self.name


class Question(core_models.UuidMixin, core_models.DescribableMixin):
    checklist = models.ForeignKey(
        to=Checklist,
        on_delete=models.CASCADE,
        related_name='questions',
    )
    order = models.PositiveIntegerField(default=0)
    category = models.ForeignKey(
        to=Category,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
    )
    solution = models.TextField(blank=True, null=True)  # It is shown when No or N/A answer is chosen

    class Meta:
        ordering = ('checklist', 'order',)

    def __str__(self):
        return self.description


class Answer(StructureModel, TimeStampedModel):
    user = models.ForeignKey(to=core_models.User, on_delete=models.SET_NULL, null=True, blank=True)
    question = models.ForeignKey(to=Question, on_delete=models.CASCADE)
    project = models.ForeignKey(to=Project, on_delete=models.CASCADE)
    value = models.NullBooleanField()

    class Permissions:
        project_path = 'project'
        customer_path = 'project__customer'



import logging
from smtplib import SMTPException

from celery import shared_task
from django.conf import settings
from django.core.mail import send_mail

logger = logging.getLogger(__name__)


@shared_task
def send_mail_notification(subject, message, to):
    try:
        send_mail(subject, message, settings.DEFAULT_FROM_EMAIL, [to])
    except SMTPException:
        message = 'Failed to send email. Receiver email: %s.' % to
        logger.warning(message)



from collections import OrderedDict

from django import template
from django.utils.lru_cache import lru_cache

from waldur_core.core.utils import get_fake_context
from waldur_core.structure import SupportedServices
from waldur_core.structure.serializers import BaseServiceSerializer

register = template.Library()


@lru_cache(maxsize=1)
@register.inclusion_tag('structure/service_settings_description.html')
def service_settings_description():
    services = []
    for cls in BaseServiceSerializer.__subclasses__():
        if cls.Meta.model is NotImplemented:
            continue
        if not SupportedServices._is_active_model(cls.Meta.model):
            continue
        name = SupportedServices.get_name_for_model(cls.Meta.model)
        fields, extra_fields = get_fields(cls)
        services.append((name, {
            'fields': fields,
            'extra_fields': extra_fields
        }))
    return {'services': sorted(services)}


def get_fields(serializer_class):
    fields = OrderedDict()
    extra_fields = OrderedDict()

    field_names = serializer_class.SERVICE_ACCOUNT_FIELDS
    if field_names is NotImplemented:
        field_names = []

    extra_field_names = serializer_class.SERVICE_ACCOUNT_EXTRA_FIELDS
    if extra_field_names is NotImplemented:
        extra_field_names = []

    serializer = serializer_class(context=get_fake_context())
    for name, field in serializer.get_fields().items():
        data = {
            'label': field.label,
            'help_text': field.help_text,
            'required': field.required
        }
        if name in field_names:
            fields[name] = data
        if name in extra_field_names:
            extra_fields[name] = data

    return (fields, extra_fields)



from django.db import migrations


def prefix_storage(apps, schema_editor):
    OfferingComponent = apps.get_model('marketplace', 'OfferingComponent')
    VolumeType = apps.get_model('openstack', 'VolumeType')
    ContentType = apps.get_model('contenttypes', 'ContentType')

    content_type = ContentType.objects.get_for_model(VolumeType)
    for offering_component in OfferingComponent.objects.filter(content_type=content_type):
        if offering_component.name.startswith('Storage ('):
            continue
        offering_component.name = 'Storage (%s)' % offering_component.name
        offering_component.save(update_fields=['name'])


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace_openstack', '0001_initial'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('openstack', '0002_volumetype'),
    ]

    operations = [
        migrations.RunPython(prefix_storage)
    ]



import django_filters

from . import models


class ProfileFilter(django_filters.FilterSet):
    user = django_filters.UUIDFilter(field_name='user__uuid')

    class Meta:
        model = models.Profile
        fields = []



from django import template
from django.conf import settings

from waldur_core.logging.middleware import get_event_context
from waldur_core.media.utils import encode_protected_url

register = template.Library()


@register.simple_tag()
def protected_url(value, field):
    if settings.DEBUG:
        return value.url
    context = get_event_context()
    return encode_protected_url(value.instance, field, user_uuid=context['user_uuid'])



from django.http.response import Http404
from rest_framework.views import APIView

from waldur_core.media.utils import send_file, get_file_from_token


class ProtectedFileView(APIView):
    authentication_classes = ()
    permission_classes = ()

    def get(self, request, token):
        value = get_file_from_token(token)
        if not value:
            raise Http404
        return send_file(value)



from django import template
from django.conf import settings

from waldur_mastermind.marketplace import plugins

from .. import models

register = template.Library()


@register.simple_tag
def get_invoice_item_component_amount(item, component):
    available_limits = plugins.manager.get_available_limits(component.component.offering.type)
    if item.limits and (
        component.component.billing_type == models.OfferingComponent.BillingTypes.USAGE or
        component.component.type in available_limits
    ):
        builtin_components = plugins.manager.get_components(component.component.offering.type)
        component_factors = {c.type: c.factor for c in builtin_components}
        factor = component_factors.get(component.component.type, 1)
        limit = item.limits.get(component.component.type, 0)
        return limit / factor

    return component.amount


@register.simple_tag
def plan_details(plan):
    context = {'plan': plan, 'components': []}

    for component in plan.components.all():
        offering_component = component.component

        if offering_component.billing_type == offering_component.BillingTypes.USAGE:
            continue

        context['components'].append({'name': offering_component.name,
                                      'amount': component.amount,
                                      'price': component.price})

    plan_template = template.Template(settings.WALDUR_MARKETPLACE.get('PLAN_TEMPLATE'))
    return plan_template.render(template.Context(context, autoescape=False))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-11-29 12:44
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0001_squashed_0054'),
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
            ],
            options={
                'verbose_name_plural': 'Categories',
            },
        ),
        migrations.CreateModel(
            name='DeploymentPlan',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('certifications', models.ManyToManyField(blank=True, to='structure.ServiceCertification')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.Project')),
            ],
            options={
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='DeploymentPlanItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('quantity', models.PositiveSmallIntegerField(default=1)),
                ('plan', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='items', to='waldur_cost_planning.DeploymentPlan')),
            ],
            options={
                'ordering': ('plan', 'preset'),
            },
        ),
        migrations.CreateModel(
            name='Preset',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('variant', models.CharField(choices=[('small', 'Small'), ('medium', 'Medium'), ('large', 'Large')], max_length=150)),
                ('ram', models.PositiveIntegerField(default=0)),
                ('cores', models.PositiveIntegerField(default=0, help_text='Preset cores count.')),
                ('storage', models.PositiveIntegerField(default=0)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='presets', to='waldur_cost_planning.Category')),
            ],
            options={
                'ordering': ('category', 'name', 'variant'),
            },
        ),
        migrations.AddField(
            model_name='deploymentplanitem',
            name='preset',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='waldur_cost_planning.Preset'),
        ),
        migrations.AlterUniqueTogether(
            name='preset',
            unique_together=set([('category', 'name', 'variant')]),
        ),
        migrations.AlterUniqueTogether(
            name='deploymentplanitem',
            unique_together=set([('plan', 'preset')]),
        ),
    ]



import django_filters
from django.contrib import auth

from waldur_core.core.filters import BaseSummaryFilterSet
from . import models

User = auth.get_user_model()


class ApplicationFilter(BaseSummaryFilterSet):
    project = django_filters.UUIDFilter(field_name='project__uuid')
    project_uuid = django_filters.UUIDFilter(field_name='project__uuid')
    project_name = django_filters.CharFilter(field_name='project__name', lookup_expr='icontains')

    class Meta:
        model = models.ApplicationModel
        fields = []



import django_filters

from waldur_core.core import filters as core_filters

from . import models


class PaymentFilter(django_filters.FilterSet):
    class Meta:
        model = models.Payment
        fields = ('customer',)

    customer = django_filters.UUIDFilter(
        field_name='customer__uuid',
        distinct=True,
    )


class InvoiceFilter(django_filters.FilterSet):
    class Meta:
        model = models.Invoice
        fields = ('customer', 'year', 'month')

    customer = core_filters.URLFilter(view_name='customer-detail', field_name='customer__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='customer__uuid')
    state = django_filters.MultipleChoiceFilter(choices=models.Invoice.States.CHOICES)



from django.conf import settings

from waldur_core.structure.permissions import _get_project
from waldur_mastermind.common import mixins as common_mixins
from waldur_mastermind.invoices import models as invoices_models
from waldur_mastermind.invoices.registrators import BaseRegistrator
from waldur_mastermind.marketplace import utils as marketplace_utils
from waldur_mastermind.packages import models as packages_models


class OpenStackItemRegistrator(BaseRegistrator):

    def get_customer(self, source):
        return source.tenant.service_project_link.project.customer

    def get_sources(self, customer):
        if not settings.WALDUR_PACKAGES['BILLING_ENABLED']:
            return packages_models.OpenStackPackage.objects.none()

        return packages_models.OpenStackPackage.objects.filter(
            tenant__service_project_link__project__customer=customer
        ).exclude(tenant__backend_id='').exclude(tenant__backend_id=None).distinct()

    def _create_item(self, source, invoice, start, end):
        package = source

        product_code = package.template.product_code
        article_code = package.template.article_code

        if package.template.unit == common_mixins.UnitPriceMixin.Units.PER_DAY:
            price = package.template.price
        else:
            price = package.template.monthly_price

        start = invoices_models.adjust_invoice_items(
            invoice, source, start, price, package.template.unit)

        item = invoices_models.InvoiceItem.objects.create(
            scope=package,
            project=_get_project(package),
            unit_price=price or 0,
            unit=package.template.unit,
            product_code=product_code,
            article_code=article_code,
            invoice=invoice,
            start=start,
            end=end,
            details=self.get_details(package))
        self.init_details(item)

    def get_details(self, source):
        package = source
        details = {
            'tenant_name': package.tenant.name,
            'tenant_uuid': package.tenant.uuid.hex,
            'template_name': package.template.name,
            'template_uuid': package.template.uuid.hex,
            'template_category': package.template.get_category_display(),
        }
        service_provider_info = marketplace_utils.get_service_provider_info(source)
        details.update(service_provider_info)
        return details

    def get_name(self, source):
        package = source
        template_category = package.template.get_category_display()
        tenant_name = package.tenant.name
        template_name = package.template.name

        if template_category:
            return '%s (%s / %s)' % (tenant_name, template_category, template_name)
        else:
            return '%s (%s)' % (tenant_name, template_name)



from waldur_rancher import models as rancher_models

from . import utils


def update_node_usage(sender, instance, created=False, **kwargs):
    if created:
        return

    utils.create_usage(instance)


def create_invoice_item_if_component_usage_has_been_created(sender, instance, created=False, **kwargs):
    component_usage = instance

    if not component_usage.tracker.has_changed('usage'):
        return

    if not isinstance(component_usage.resource.scope, rancher_models.Cluster):
        return

    utils.component_usage_register(component_usage)
